import os
import pdb
import sys

from stoobly_agent.app.cli.helpers.handle_replay_service import DEFAULT_FORMAT, handle_before_replay, handle_after_replay, print_session, ReplaySession
from stoobly_agent.app.cli.helpers.handle_test_service import SessionContext, exit_on_failure, handle_test_complete, handle_test_session_complete 
from stoobly_agent.app.cli.helpers.print_service import print_scenarios, select_print_options
from stoobly_agent.app.cli.helpers.test_facade import TestFacade
from stoobly_agent.app.cli.helpers.context import ReplayContext
from stoobly_agent.app.models.helpers.apply import Apply
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import alias_resolve_strategy, env_vars
from stoobly_agent.lib import logger

from ..helpers.scenario_facade import ScenarioFacade
from ..helpers.validations import *
from ..types.scenario import ScenarioTestOptions

def create_handler(kwargs):
  print_options = select_print_options(kwargs)

  settings = Settings.instance()
  project_key = resolve_project_key_and_validate(kwargs, settings) 

  scenario = ScenarioFacade(settings)

  res, status = scenario.create(project_key, kwargs['name'], kwargs['description'])

  if filter_response(res, status):
      sys.exit(1)

  print_scenarios([res], **print_options)

def delete_handler(kwargs):
  validate_scenario_key(kwargs['scenario_key'])

  scenario = ScenarioFacade(Settings.instance())
  res, status = scenario.show(kwargs['scenario_key'])

  if filter_response(res, status):
    sys.exit(1)

  is_deleted = res.get('is_deleted')
  if not res:
    print(f"Error: Could not find scenario", file=sys.stderr)
    sys.exit(1)

  res = scenario.delete(kwargs['scenario_key'])
  if not res:
    print('Error: Could not delete scenario', file=sys.stderr)
    sys.exit(1)

  if not is_deleted:
    print('Scenario moved to trash!')
  else:
    print('Scenario deleted!')

def list_handler(kwargs):
  print_options = select_print_options(kwargs)

  settings = Settings.instance()

  project_key = resolve_project_key_and_validate(kwargs, settings)
  if 'project_key' in kwargs:
      del kwargs['project_key']

  if kwargs.get('search'):
      kwargs['q'] = kwargs['search']
      del kwargs['search']

  scenario = ScenarioFacade(settings)

  scenarios_response, status = scenario.index(project_key, kwargs)
  if filter_response(scenarios_response, status):
      sys.exit(1)

  if len(scenarios_response['list']) == 0:
      print('No scenarios found.')
  else:
      print_scenarios(scenarios_response['list'], **print_options) 

def replay_handler(kwargs):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']

  scenario_key = validate_scenario_key(kwargs['key'])

  if kwargs.get('scenario_key'):
      if kwargs['scenario_key'] and not kwargs.get('record'):
          print("Error: Missing option '--record'.", file=sys.stderr)
          sys.exit(1)

      validate_scenario_key(kwargs['scenario_key'])

  scenario = ScenarioFacade(Settings.instance())

  scenario_response, status = scenario.show(kwargs.get('key'))
  if filter_response(scenario_response, status):
      sys.exit(1)

  if 'validate' in kwargs and len(kwargs['validate']):
      validate_aliases(kwargs['validate'], assign=kwargs['assign'], format=kwargs['format'], trace_id=kwargs['trace_id'])

  __assign_default_alias_resolve_strategy(kwargs)
  __assign_default_origin(kwargs)

  session: ReplaySession = {
      'buffer': kwargs['format'] and kwargs['format'] != DEFAULT_FORMAT,
      'contexts': [],
      'format': kwargs['format'],
      'scenario_id': scenario_key.id,
      'total': 0,
  }

  kwargs['before_replay'] = lambda context: handle_before_replay(
    context, session
  )
  kwargs['after_replay'] = lambda context: handle_after_replay(
      context, session
  )

  scenario.replay(kwargs.get('key'), kwargs)

  print_session(session)

def reset_handler(kwargs):
  scenario_key = kwargs['scenario_key']
  _scenario_key = validate_scenario_key(scenario_key)

  apply_service = Apply(force=kwargs['force']).with_logger(print)
  resetted = apply_service.scenario(_scenario_key.id)

  if not resetted:
    print('Successfully reset the scenario!')
  else:
    print('Could not reset the scenario.')

def show_handler(kwargs):
  print_options = select_print_options(kwargs)

  settings = Settings.instance()
  scenario_key = resolve_scenario_key_and_validate(kwargs, settings)
  scenario = ScenarioFacade(settings)

  scenario_response, status = scenario.show(scenario_key)
  if filter_response(scenario_response, status):
      sys.exit(1)

  print_scenarios([scenario_response], **print_options)

def snapshot_handler(kwargs):
  scenario_key = kwargs['scenario_key']
  del kwargs['scenario_key']

  validate_scenario_key(scenario_key)

  scenario = ScenarioFacade(Settings.instance())
  scenario.snapshot(scenario_key, kwargs)

def test_handler(kwargs: ScenarioTestOptions):
  os.environ[env_vars.LOG_LEVEL] = kwargs['log_level']
  logger.Logger.reload()

  settings = Settings.instance()
  scenario_key = validate_scenario_key(kwargs['key'])

  if kwargs.get('remote_project_key'):
      validate_project_key(kwargs['remote_project_key'])

  if kwargs.get('report_key'):
      validate_report_key(kwargs['report_key'])
      kwargs['save'] = True # If report_key is set, then intention is to save results to the report

  if len(kwargs['validate']):
      validate_aliases(kwargs['validate'], assign=kwargs['assign'], format=kwargs['format'], trace_id=kwargs['trace_id'])

  __assign_default_alias_resolve_strategy(kwargs)
  __assign_default_origin(kwargs)

  session: ReplaySession = {
      'buffer': kwargs['format'] and kwargs['format'] != DEFAULT_FORMAT,
      'contexts': [],
      'format': kwargs['format'],
      'scenario_id': scenario_key.id,
      'total': 0,
  }

  session_context: SessionContext = { 
      'aggregate_failures': kwargs['aggregate_failures'], 
      'passed': 0, 
      'project_id': scenario_key.project_id,
      'skipped': 0, 
      'test_facade': TestFacade(settings), 
      'total': 0 
  }

  kwargs['before_replay'] = lambda context: handle_before_replay(
      context, session,
  )
  kwargs['after_replay'] = lambda context: __handle_on_test_response(
      context, session_context, kwargs
  )

  scenario = ScenarioFacade(settings)
  scenario.test(kwargs['key'], kwargs)

  handle_test_session_complete(session_context, format=kwargs['format'])

  exit_on_failure(session_context, format=kwargs['format'])

def __handle_on_test_response(replay_context: ReplayContext, session_context: SessionContext, kwargs):
    handle_test_complete(replay_context, session_context, format=kwargs['format'], output_level=kwargs['output_level'])

    if not session_context['aggregate_failures']:
        exit_on_failure(session_context, complete=False, format=kwargs['format'])

def __assign_default_alias_resolve_strategy(kwargs):
    # If we have assigned values to aliases, it's likely we want to also have them resolved
    if 'assign' in kwargs and len(kwargs['assign']) > 0 and kwargs['alias_resolve_strategy'] == alias_resolve_strategy.NONE:
        kwargs['alias_resolve_strategy'] = alias_resolve_strategy.FIFO

def __assign_default_origin(kwargs):
    if not kwargs.get('host'):
        kwargs['host'] = os.environ.get(env_vars.AGENT_REPLAY_HOST)

    if not kwargs.get('scheme'):
        kwargs['scheme'] = os.environ.get(env_vars.AGENT_REPLAY_SCHEME)


AFTER_MOCK = 'handle_after_mock'
AFTER_RECORD = 'handle_after_record'
AFTER_REPLAY = 'handle_after_replay'
AFTER_TEST = 'handle_after_test'
BEFORE_MOCK = 'handle_before_mock'
BEFORE_RECORD = 'handle_before_record'
BEFORE_REPLAY = 'handle_before_replay'
BEFORE_REQUEST = 'handle_before_request'
BEFORE_RESPONSE = 'handle_before_response'
BEFORE_TEST = 'handle_before_test'
ON_LENGTH_MATCH_ERROR = 'handle_length_match_error'
ON_PARAM_NAME_EXISTS_ERROR = 'handle_param_name_exists_error'
ON_PARAM_NAME_MISSING_ERROR = 'handle_param_name_missing_error'
ON_TYPE_MATCH_ERROR = 'handle_type_match_error'
ON_VALID_TYPE_ERROR = 'handle_valid_type_error'
ON_VALUE_MATCH_ERROR = 'handle_value_match_error'

BEFORE_REQUEST_COMPONENT_CREATE = 'handle_before_request_component_create'
BEFORE_REQUEST_COMPONENT_DELETE = 'handle_before_request_component_delete'


import pdb

from mitmproxy.http import Request as MitmproxyRequest
from typing import Union

from ..replay.body_parser_service import decode_response, encode_response

class MitmproxyRequestBodyFacade:
    def __init__(self, request: MitmproxyRequest):
        self.__request = request

    def get(self, content_type: Union[bytes, str]):
        return decode_response(self.__request.content, content_type)

    def set(self, content, content_type: Union[bytes, str]):
        """
        Adjusting Content-Length header should be done by MitmproxyRequest
        """
        self.__request.content = encode_response(content, content_type).encode()


from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from typing import List, Optional, TypedDict

from stoobly_agent.app.proxy.record.joined_request import JoinedRequest

class RequestCreateParams(TypedDict):
  flow: MitmproxyHTTPFlow
  project_id: str
  joined_request: JoinedRequest
  scenario_id: str
  uuid: str

class RequestDestroyParams(TypedDict):
  force: bool

class RequestDestroyAllParams(TypedDict):
  ids: List[int] 
  scenario_id: int

class RequestIndexSimilarParams(TypedDict):
  host: str
  port: str
  method: str
  pattern: str
  scenario_id: Optional[int]

class RequestShowParams(TypedDict):
  body: bool
  headers: bool
  project_id: str
  query_params: bool
  response: bool


import json
import pdb
import pytest

from click.testing import CliRunner
from mock import patch
from typing import Tuple
from unittest.mock import MagicMock

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.cli import record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.lib.orm.scenario import Scenario

from stoobly_agent.app.models.factories.resource.local_db.request_adapter import LocalDBRequestAdapter
from stoobly_agent.app.proxy.replay.replay_request_service import ReplayRequestOptions, replay
from stoobly_agent.app.proxy.record.upload_request_service import upload_request

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestReplayIntegration():
    @pytest.fixture(scope='class', autouse=True)
    def settings(self):
        return reset()

    class TestWhenNoOptions():

        @pytest.fixture(scope='class')
        def created_scenario(self, runner: CliRunner):
            create_result = runner.invoke(scenario, ['create', 'test'])
            assert create_result.exit_code == 0
            return Scenario.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_one(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class', autouse=True)
        def recorded_request_two(self, runner: CliRunner, created_scenario: Scenario):
            record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
            assert record_result.exit_code == 0
            return Request.last()

        @pytest.fixture(scope='class')
        def replay_spy(self, runner: CliRunner, created_scenario: Scenario):
            with patch('stoobly_agent.app.proxy.replay.replay_request_service.replay') as test_patch:
                replay_result = runner.invoke(scenario, ['replay', created_scenario.key()])
                assert replay_result.exit_code == 0
                return test_patch

        @pytest.fixture(scope='class')
        def replay_options(self, replay_spy: MagicMock) -> ReplayRequestOptions:
            return replay_spy.call_args.args[1]
            
        def test_it_called_twice(self, replay_spy: MagicMock):
            assert replay_spy.call_count == 2

        def test_options_has_project_key(self, replay_options: ReplayRequestOptions):
            assert replay_options.get('project_key') is not None

        def test_options_not_has_scenario_key(self, replay_options: ReplayRequestOptions):
            assert replay_options.get('scenario_key') is None

    class TestWhenRecord():

        class TestWhenScenario():
            @pytest.fixture(scope='class', autouse=True)
            def settings(self):
                return reset()

            @pytest.fixture(scope='class')
            def source_scenario(self, runner: CliRunner):
                create_result = runner.invoke(scenario, ['create', 'test'])
                assert create_result.exit_code == 0
                return Scenario.last()

            @pytest.fixture(scope='class')
            def destination_scenario(self, runner: CliRunner):
                create_result = runner.invoke(scenario, ['create', 'test'])
                assert create_result.exit_code == 0
                return Scenario.last()

            @pytest.fixture(scope='class', autouse=True)
            def recorded_request(self, runner: CliRunner, source_scenario: Scenario):
                record_result = runner.invoke(record, ['--scenario-key', source_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
                assert record_result.exit_code == 0
                return Request.last()

            @pytest.fixture(scope='class')
            def spies(self, runner: CliRunner, source_scenario: Scenario, destination_scenario: Scenario):
                @patch.object(LocalDBRequestAdapter, 'create', wraps=LocalDBRequestAdapter(Request, Response).create)
                @patch('stoobly_agent.app.proxy.record.upload_request_service.upload_request', wraps=upload_request)
                @patch('stoobly_agent.app.proxy.replay.replay_request_service.replay', wraps=replay)
                def spy_on(replay_spy: MagicMock, upload_request_spy: MagicMock, create_spy: MagicMock):
                    replay_result = runner.invoke(scenario, ['replay', '--record' , '--scenario-key', destination_scenario.key(), source_scenario.key()])
                    assert replay_result.exit_code == 0
                    return replay_spy, upload_request_spy, create_spy

                return spy_on()

            @pytest.fixture(scope='class')
            def replay_spy(self, spies: Tuple[MagicMock]):
                return spies[0]

            @pytest.fixture(scope='class')
            def replay_options(self, replay_spy: MagicMock) -> ReplayRequestOptions:
                return replay_spy.call_args.args[1]

            @pytest.fixture(scope='class')
            def upload_request_spy(self, spies: Tuple[MagicMock]):
                return spies[1]

            @pytest.fixture(scope='class')
            def upload_request_intercept_settings(self, upload_request_spy: MagicMock):
                return upload_request_spy.call_args.args[1]

            @pytest.fixture(scope='class')
            def create_spy(self, spies: Tuple[MagicMock]):
                return spies[2]

            @pytest.fixture(scope='class')
            def create_kwargs(self, create_spy: MagicMock):
                return create_spy.call_args.kwargs

            def test_replay_called_once(self, replay_spy: MagicMock):
                replay_spy.assert_called_once()

            def test_options_has_project_key(self, replay_options: ReplayRequestOptions):
                assert replay_options.get('project_key') is not None

            def test_options_not_has_scenario_key(self, replay_options: ReplayRequestOptions, destination_scenario: Scenario):
                assert replay_options.get('scenario_key') == destination_scenario.key()

            def test_upload_request_called_once(self, upload_request_spy: MagicMock):
                upload_request_spy.assert_called_once()

            def test_upload_request_called_with_destination_scenario_key(self, upload_request_intercept_settings, destination_scenario: Scenario):
                assert upload_request_intercept_settings.scenario_key == destination_scenario.key()

            def test_create_called_once(self, create_spy: MagicMock):
                create_spy.assert_called_once()

            def test_create_called_with_destination_scenario_id(self, create_kwargs, destination_scenario: Scenario):
                assert create_kwargs['scenario_id'] == destination_scenario.uuid



from .config import Config
from .constants import APP_NAME_ENV, APP_NETWORK_ENV

class AppConfig(Config):

  def __init__(self, dir: str):
    super().__init__(dir)

    self.__name = None
    self.__network = None

    self.load()

  @property
  def name(self):
    return self.__name

  @name.setter
  def name(self, v):
    self.__name = v

  @property
  def network(self):
    return self.__network

  @network.setter
  def network(self, v):
    self.__network = v

  def load(self, config = None):
    config = config or self.read()

    self.name = config.get(APP_NAME_ENV)
    self.network = config.get(APP_NETWORK_ENV)
    
  def write(self):
    config = {}

    if self.name:
      config[APP_NAME_ENV] = self.name

    if self.network:
      config[APP_NETWORK_ENV] = self.network

    super().write(config)


import pdb
import requests

from stoobly_agent.lib.utils.decode import decode

CRLF = b'\r\n'
DEFAULT_HTTP_VERSION = b'HTTP/1.1'

HEADER_COOKIE = 'cookie'
HEADER_COOKIE_SEPARATOR = '; '
HEADER_SEPARATOR = ', '

class RawHttpRequestAdapter():

  def __init__(self, req_text):
    req_lines = req_text.split(CRLF)
    self.__parse_request_line(req_lines[0])

    ind = 1
    self.headers = dict()
    while ind < len(req_lines) and len(req_lines[ind]) > 0:
        colon_ind = req_lines[ind].find(b':')
        header_key = req_lines[ind][:colon_ind]
        header_value = req_lines[ind][colon_ind + 1:]

        decoded_key = decode(header_key)
        decoded_value = decode(header_value).strip()

        if decoded_key in self.headers:
          separator = HEADER_COOKIE_SEPARATOR if decoded_key.lower() == HEADER_COOKIE else HEADER_SEPARATOR
          self.headers[decoded_key] = separator.join([self.headers[decoded_key], decoded_value])
        else:
          self.headers[decoded_key] = decoded_value

        ind += 1

    ind += 1
    data_lines = req_lines[ind:] if ind < len(req_lines) else None
    self.body = CRLF.join(data_lines)

  def to_request(self) -> requests.Request:
    req = requests.Request(
      method=self.method,
      url=self.url,
      headers=self.headers,
      data=self.body
    )

    return req

  def __parse_request_line(self, request_line):
    request_parts = request_line.split(b' ')
    self.method = decode(request_parts[0])
    self.url = decode(request_parts[1])
    self.protocol = decode(request_parts[2]) if len(request_parts) > 2 else DEFAULT_HTTP_VERSION

  def __str__(self):
    headers = CRLF.join(f'{key}: {self.headers[key]}' for key in self.headers)
    return f'{self.method} {self.url} {self.protocol}{CRLF}' \
            f'{headers}{CRLF}{CRLF}{self.body}'




from .cli_settings import *
from .proxy_settings import *
from .remote_settings import *
from .ui_settings import *

from typing import List, Optional, TypedDict, Union

class Ignore(TypedDict):
    name: str
    type: str

class Redact(TypedDict):
    name: str
    type: str

class Rewrite(TypedDict):
    name: Optional[str]
    type: str
    value: str

class IgnoreRule(TypedDict):
    ignores: List[Ignore]
    method: str
    pattern: str

class RewriteRule(TypedDict):
    rewrites: List[Rewrite]
    method: str
    pattern: str

class IProjectRecordSettings(TypedDict):
    enabled: bool
    exclude_patterns: list
    rewrite_rules: list
    include_patterns: list
    policy: str
    project_key: str
    service_url: str
    scenario_key: str

class IProjectMockSettings(TypedDict):
    enabled: bool
    exclude_patterns: list
    include_patterns: list
    policy: str
    project_key: str
    service_url: str
    scenario_key: str

class IProjectTestSettings(TypedDict):
    exclude_patterns: list
    include_patterns: list
    policy: str
    project_key: str
    rewrite_rules: list
    service_url: str
    scenario_key: str
    test_strategy: str
    enabled: bool

class IRecordSettings(TypedDict):
    enabled: bool
    project_key: str
    settings: dict

class IMockSettings(TypedDict):
    enabled: bool
    project_key: str
    settings: dict

class ITestSettings(TypedDict):
    enabled: bool
    project_key: str
    settings: dict

class ISettingsMode(TypedDict):
    active: str
    record: IRecordSettings
    mock: IMockSettings
    test: ITestSettings

class ISettings(TypedDict):
    agent_url: str
    api_url: str
    api_key: str
    mode: ISettingsMode
    proxy_config_path: str

Component = {
    'Header': 'Header',
    'BodyParam': 'Body Param',
    'QueryParam': 'Query Param',
}
IProjectModeSettings = Union[IProjectMockSettings, IProjectRecordSettings, IProjectTestSettings]
Rule = Union[IgnoreRule, RewriteRule]


import os

from .constants import ENV_FILE

DELIMITTER = "\n"

class Env():

  def __init__(self, path: str):
    self.__path = path

  @property
  def path(self):
    return self.__path

  def read(self, source: str = None):
    if not source:
      source = self.path

    env_vars = {}

    if os.path.exists(source):
      with open(source, 'r') as fp:
        contents = fp.read()

        rows = contents.split(DELIMITTER)
        for row in rows:
          el = row.split("=")
          if len(el) != 2:
            continue
          env_vars[el[0]] = el[1]

    return env_vars

  # Priority: env_vars, source_env_vars, dest_env_vars
  def merge(self, defaults_path: str, env_vars: dict = {}):
    source_env_vars = self.read()
    dest_env_vars = self.read(defaults_path)
    dest_env_vars.update(source_env_vars)
    dest_env_vars.update(env_vars)
    return env_vars

  def write(self, env_vars: dict):
    env_file_path = os.path.join(self.path)
    with open(env_file_path, 'w') as fp:
      contents = []
      for var in env_vars:
          contents.append(f"{var}={env_vars[var]}")
      fp.write(DELIMITTER.join(contents))


import pdb
import uuid

from stoobly_orator.migrations import Migration

from stoobly_agent.app.proxy.record.request_string_control import RequestStringControl
from stoobly_agent.lib.orm.request import Request

class AddUuidColumnToRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.table('requests') as table:
            table.string('uuid', 36).default('').index().unique()

        for request in Request.all():
            control = RequestStringControl(request.control)
            try:
                request.update(uuid=str(uuid.UUID(control.id)))
            except Exception:
                request.update(uuid=str(uuid.uuid4()))

    def down(self):
        """
        Revert the migrations.
        """
        with self.schema.table('requests') as table:
            table.drop_column('uuid')



import pdb

from typing import List, Tuple, TypedDict, Union

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.lib.api.interfaces.endpoints import RequestComponentName

from ..context_abc import TestContextABC as TestContext
from .context import MatchContext, build_match_context
from .errors import length_match_error, param_name_exists_error, param_name_missing_error, type_match_error, valid_type_error

class Options(TypedDict):
    strict: bool # Expect actual to not have fields not specified in endpoint spec

def matches(test_context: TestContext, facade: RequestComponentNamesFacade, actual, **options: Options):
    context = build_match_context(test_context, facade)

    if context.value_is_dict(actual):
        return dict_matches(context, actual, **options)
    elif context.value_is_list(actual):
        return list_matches(context, actual, **options)
    else:
        # TODO: handle non-traversible content
        return True, ''

def dict_matches(parent_context: MatchContext, actual: dict, **options: Options) -> Tuple[bool, str]:
    if not parent_context.value_is_dict(actual):
        if not parent_context.handle_type_match_error(actual):
            return type_match_error(parent_context.path_key, dict, type(actual))

    component_names = parent_context.children

    # Check all properties required by contract exist
    for component_name in component_names:
        key = component_name['name']
        if not parent_context.param_name_exists(key, actual):
            context = MatchContext(parent_context.to_dict())
            context.visit_dict(key)

            if context.ignored():
                continue

            if not context.handle_param_name_missing_error(actual):
                return param_name_missing_error(context.path_key)

    index = parent_context.request_component_names_query_index

    for key in actual.copy(): # Create a copy, actual might change during loop
        actual_value = actual[key]
        context = MatchContext(parent_context.to_dict())
        context.visit_dict(key)

        path_key = context.path_key
        query = context.query

        # Check if the key exists in the contract
        if not context.param_name_exists(query, index):
            if not context.handle_param_name_exists_error(actual):
                if not options.get('strict'):
                    continue
                else:
                    return param_name_exists_error(path_key)
        
        contracts = index.get(query)
        if not contracts:
            continue

        if not context.value_contract_matches(actual_value, contracts):
            if not context.handle_type_match_error(actual):
                if context.ignored():
                    continue
                
                valid_types = __build_valid_types(contracts)
                return valid_type_error(path_key, actual_value, valid_types)

        if context.value_is_dict(actual_value):
            matches, log = dict_matches(context, actual_value, **options)
            if not matches:
                return matches, log
        elif context.value_is_list(actual_value):
            matches, log = list_matches(context, actual_value, **options)
            if not matches:
                return matches, log

    return True, ''

###
#
# For list matching, look for all types in expected list
# Expect tested list to have a type that exists in expected list
#
# If list value is traversable, traverse
#
def list_matches(parent_context: MatchContext, actual: list, **options: Options) -> Tuple[bool, str]:
    parent_path_key = f"{parent_context.path_key}[*]"
    if not parent_context.value_is_list(actual):
        if not parent_context.handle_type_match_error(actual):
            return type_match_error(parent_path_key, list, type(actual))

    index = parent_context.request_component_names_query_index
    contracts: List[RequestComponentName] = index.get(parent_path_key) or []

    # If required, make sure not empty 
    if parent_context.required and len(actual) == 0 and len(contracts) > 0:
        if not parent_context.handle_length_match_error(actual):
            return length_match_error(parent_path_key, None, actual) 

    valid_types = __build_valid_types(contracts)

    for i in range(len(actual)):
        value = actual[i]
        context = MatchContext(parent_context.to_dict())
        context.visit_list(i)
        path_key = context.path_key

        if not context.value_contract_matches(value, contracts):
            if not context.handle_valid_type_error(actual):
                if not context.selected() or not context.required():
                    continue
                
                return valid_type_error(path_key, value, valid_types)

        if context.value_is_dict(value):
            matches, log = dict_matches(context, value, **options)
            if not matches:
                return matches, log
        elif context.value_is_list(value):
            matches, log = list_matches(context, value, **options)
            if not matches:
                return matches, log

    return True, ''

def __build_valid_types(contracts: Union[None, RequestComponentName, List[RequestComponentName]]):
    valid_types = []
    if contracts:
        if not isinstance(contracts, list):
            contracts = [contracts]

        valid_types = list(map(lambda contract: contract.get('inferred_type'), contracts))
    return valid_types


import uuid

from .resource_key import ResourceKey

DELIMITTER = '.'

class UuidKey(ResourceKey):
  def __init__(self, key: str):
    super().__init__(super().encode(self.__decode(key)))

  @staticmethod
  def encode(d: dict):
    toks = []
    for key in d:
      toks.append(f"{key}{d[key]}")

    return DELIMITTER.join(toks)

  @property
  def id(self) -> str:
    try:
      u = uuid.UUID(self.get('i'))
    except TypeError:
      return ''
    return str(u)

  def __decode(self, key: str):
    toks = key.split(DELIMITTER)

    d = {}
    for tok in toks:
      d[tok[0]] = tok[1:]

    return d


from lib.api.interfaces.requests import RequestShowResponse

from .mitmproxy_adapter import MitmproxyRequestAdapter

class StooblyRequestAdapterFactory():

  def __init__(self, request: RequestShowResponse):
    self.__request = request

  def mitmproxy_request(self):
    return MitmproxyRequestAdapter(self.__request)


import pdb
import requests

from stoobly_agent.app.api.simple_http_request_handler import SimpleHTTPRequestHandler
from stoobly_agent.app.models.replayed_response_model import ReplayedResponseModel
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import custom_headers

class ReplayedResponsesController:
    _instance = None

    def __init__(self):
        if self._instance:
            raise RuntimeError('Call instance() instead')
        else:
            self.data = {}

    @classmethod
    def instance(cls):
        if cls._instance is None:
            cls._instance = cls()

        return cls._instance

    # GET /requests/:requestId/replayed_responses
    def index(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'request_id': 1
        })
        replayed_response_model = self.__replayed_response_model(context)
        replayed_responses, status = replayed_response_model.index(**context.params)

        if context.filter_response(replayed_responses, status):
            return

        context.render(
            json = replayed_responses,
            status = 200
        )

    # GET /requests/:requestId/replayed_responses/:replayedResponseId/mock
    def mock(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'request_id': 1,
            'replayed_response_id': 3,
        })

        replayed_response_model = self.__replayed_response_model(context) 
        response, status = replayed_response_model.mock(context.params.get('replayed_response_id'))

        if context.filter_response(response, status):
            return

        return self.__render_response(context, response)

    # PUT /requests/:requestId/replayed_responses/:replayedResponseId/activate
    def activate(self, context: SimpleHTTPRequestHandler):
        context.parse_path_params({
            'request_id': 1,
            'replayed_response_id': 3,
        })

        replayed_response_model = self.__replayed_response_model(context) 
        replayed_response, status = replayed_response_model.activate(context.params.get('replayed_response_id'))

        if context.filter_response(replayed_response, status):
            return

        context.render(
            json = replayed_response,
            status = 200
        )

    def __replayed_response_model(self, context: SimpleHTTPRequestHandler):
        access_token = context.headers.get('access-token')
        replayed_response_model = ReplayedResponseModel(Settings.instance(), access_token=access_token)
        return replayed_response_model

    def __render_response(self, context: SimpleHTTPRequestHandler, response: requests.Response):
        if response == None:
            return context.render(
                plain = '',
                status = 404
            )

        # Extract specific headers
        headers = {
            'Access-Control-Expose-Headers': custom_headers.RESPONSE_ID,
        }

        accepted_headers = ['content-type', custom_headers.RESPONSE_ID.lower()]

        for header, val in response.headers.items():
            decoded_header = header.lower()

            if decoded_header not in accepted_headers:
                continue 

            headers[decoded_header] = val

        context.render(
            data = response.content,
            headers = headers,
            status = 200
        )


import requests
import urllib

from stoobly_agent.lib.api.interfaces.tests import TestCreateParams

from ..logger import Logger
from .stoobly_api import StooblyApi

LOG_ID = 'TestsResource'

class TestsResource(StooblyApi):
  TESTS_ENDPOINT = 'tests'

  def create(self, project_id: str, raw_request, params: TestCreateParams = {}):
    url = f"{self.service_url}/{self.TESTS_ENDPOINT}"

    body = {
        'project_id': project_id,
        **params,
    }

    return self.post(url, headers=self.default_headers, data=body, files={ 'request': raw_request })

  def show(self, test_id: int, **query_params) -> requests.Response:
    url = f"{self.service_url}/{self.TESTS_ENDPOINT}/{test_id}"

    Logger.instance(LOG_ID).debug(f"{url}?{urllib.parse.urlencode(query_params)}")

    return self.get(url, headers=self.default_headers, params=query_params)


import pdb

from typing import Iterable, Tuple
from stoobly_agent.app.proxy.test.context_abc import TestContextABC as TestContext
from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.test.matchers.errors import param_name_missing_error, type_match_error, valid_type_error

from .context import MatchContext, build_match_context

def matches(test_context: TestContext, facade: RequestComponentNamesFacade, expected, actual):
    context = build_match_context(test_context, facade)

    if not context.value_type_matches(expected, actual):
        return type_match_error(context.path_key, expected, actual)

    if context.value_is_dict(expected):
        return dict_matches(context, expected, actual)
    elif context.value_is_list(expected):
        return list_matches(context, expected, actual)
    else:
        return value_matches(context, expected, actual)

def value_matches(parent_context: MatchContext, expected, actual):
    if not parent_context.value_fuzzy_matches(expected, actual):
        return type_match_error(parent_context.path_key, type(expected), (actual))
    return True, ''

def dict_matches(parent_context: MatchContext, expected: dict, actual: dict) -> Tuple[bool, str]:
    if not parent_context.value_is_dict(expected) or not parent_context.value_is_dict(actual):
        return type_match_error(parent_context.path_key, type(expected), type(actual))

    for key, expected_value in expected.items():
        context = MatchContext(parent_context.to_dict())
        context.visit_dict(key)
        path_key = context.path_key

        if not context.param_name_exists(key, actual):
            if not context.selected() or not context.required():
                continue 

            return param_name_missing_error(path_key)

        actual_value = actual[key]
        if not context.value_fuzzy_matches(actual_value, expected_value):
            if context.ignored(expected_value, actual_value):
                continue

            return type_match_error(path_key, type(expected_value), type(actual_value))

        if context.value_is_dict(actual_value):
            matches, log = dict_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log
        elif context.value_is_list(actual_value):
            matches, log = list_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log

    return True, ''

###
#
# For list matching, look for all types in expected list
# Expect tested list to have a type that exists in expected list
#
# If list value is traversable, traverse
#
def list_matches(parent_context: MatchContext, expected: list, actual: list) -> Tuple[bool, str]:
    if not parent_context.value_is_list(expected) or not parent_context.value_is_list(actual):
        return type_match_error(parent_context.path_key, type(expected), type(actual))

    valid_types = []
    type_examples = {}
    for i, value in enumerate(expected):
        valid_types.append(type(value))

        if type(value) is dict and str(dict) not in type_examples:
            type_examples[dict] = value
        elif type(value) is list and str(list) not in type_examples:
            type_examples[list] = value

    for i, value in enumerate(actual):
        context = MatchContext(parent_context.to_dict())
        context.visit_list(i)
        path_key = context.path_key

        if not context.value_type_exists(value, valid_types):
            if not context.selected() or not context.required():
                continue

            return valid_type_error(path_key, value, valid_types)

        if context.value_is_dict(value):
            matches, log = dict_matches(context, type_examples[dict], value)
            if not matches:
                return matches, log
        elif context.value_is_list(value):
            matches, log = list_matches(context, type_examples[list], value)
            if not matches:
                return matches, log

    return True, ''


import copy
import itertools
import re
import yaml

from functools import reduce
from jsonschema_path import SchemaPath
from openapi_core import OpenAPI
from pprint import pprint
from typing import Dict, List, Union
from urllib.parse import urlparse

from stoobly_agent.lib.api.interfaces.endpoints import (
  Alias,
  EndpointShowResponse,
  RequestComponentName,
)
from stoobly_agent.lib.utils.python_to_ruby_type import convert_reverse

from .schema_builder import SchemaBuilder

class OpenApiEndpointAdapter():
  def __init__(self):
    return

  def adapt_from_file(self, file_path) -> List[EndpointShowResponse]:
    spec = {}

    with open(file_path, "r") as stream:
      file_data: Dict = yaml.safe_load(stream)

      if 'info' not in file_data:
        self.__add_info(file_data)

      missing_oauth2_scopes = self.__is_missing_oauth2_scopes(file_data)
      if missing_oauth2_scopes:
        self.__add_oauth2_default_scopes(file_data)

      openApi = OpenAPI.from_dict(file_data)
      spec = openApi.spec

    return self.adapt(spec)

  def adapt(self, spec: SchemaPath) -> List[EndpointShowResponse]:
    endpoints = []
    endpoint_counter = 0
    components = spec.get("components", {})
    schemas = components.get("schemas", {})
    paths = spec.getkey('paths')

    servers_spec = spec.get("servers")
    servers = self.__evaluate_servers(servers_spec)

    for _, server in enumerate(servers):
      url = server["url"]

      for path_name, path in paths.items():
        operations = [
            "get",
            "put",
            "post",
            "delete",
            "options",
            "head",
            "patch",
            "trace",
        ]
        for http_method in operations:
          if http_method not in path:
            continue

          endpoint_counter += 1

          parsed_url = urlparse(url)
          endpoint: EndpointShowResponse = {}
          endpoint['id'] = endpoint_counter
          endpoint['method'] = http_method.upper()
          endpoint['host'] = '-' if parsed_url.netloc == '' else parsed_url.netloc

          joined_path = self.__urljoin(parsed_url.path, path_name)
          split_parts = joined_path.split('/')
          pattern_path = []
          segment_names = []
          for part in split_parts:
            sanitized_part = part
            segment_name = part
            if part.startswith('{') and part.endswith('}'):
              sanitized_part = '%' 
              segment_name = f":{part[1:-1]}"
            pattern_path.append(sanitized_part)
            segment_names.append(segment_name)
          pattern_path_str = '/'.join(pattern_path)
          endpoint['match_pattern'] = pattern_path_str
          endpoint['path'] = joined_path

          endpoint['path_segment_names'] = []
          for segment_name in segment_names:
            if segment_name == "":
              continue
            path_component_name: RequestComponentName = {}
            path_component_name['name'] = segment_name
            path_component_name['type'] = "alias" if segment_name.startswith(':') else "static"
            endpoint['path_segment_names'].append(path_component_name)
          
          endpoint['port'] = str(parsed_url.port)
          if endpoint['port'] is None or endpoint['port'] == 'None':
            if parsed_url.scheme == 'https':
              endpoint['port'] = '443'
            elif parsed_url.scheme == 'http':
              endpoint['port'] = '80'
            else:
              endpoint['port'] = '0'

          alias_counter = 0
          header_param_counter = 0
          operation = path[http_method]
          required_query_params = []
          required_body_params = []
          literal_query_params = []

          parameters = operation.get("parameters", {})
          for parameter in parameters:
            if '$ref' in parameter.keys():
              parameter = self.__dereference(components, parameter['$ref'])

            if parameter['in'] == 'query':
              # query_param: RequestComponentName = {}
              # query_param['name'] = parameter['name']
              #
              # query_param_example = parameter.get('example')
              # if query_param_example:
              #   query_param['values'].append(query_param_example)

              if parameter.get('required') == True:
                required_query_params.append(parameter['name'])

              if not endpoint.get('query_param_names'):
                endpoint['query_param_names'] = []

              schema = parameter.get('schema', {})
              self.__extract_param_properties(components, required_query_params, {parameter['name']: schema}, literal_query_params, curr_id=len(literal_query_params)+1)

              endpoint['literal_query_params'] = literal_query_params

            elif parameter['in'] == 'header':
              header: RequestComponentName = {}
              header['name'] = parameter['name']
              header_example = parameter.get('example')
              if header_example:
                header['values'].append(header_example)
              if parameter['required'] == True:
                header['is_required'] = True
              else:
                header['is_required'] = False

              if not endpoint.get('header_names'):
                endpoint['header_names'] = []
              header['is_deterministic'] = True
              header_param_counter += 1
              header['id'] = header_param_counter

              endpoint['header_names'].append(header)

            elif parameter['in'] == 'path':
              if not endpoint.get('aliases'):
                endpoint['aliases'] = []

              found_alias = None
              for alias in endpoint['aliases']:
                if parameter['name'] == alias['name']:
                  found_alias = alias

              if not found_alias:
                alias: Alias = {}
                alias_counter += 1
                alias['id'] = alias_counter
                alias['name'] = '{' + parameter['name'] + '}'

                endpoint['aliases'].append(alias)

          # TODO: nested servers

          request_body = operation.get("requestBody", {})
          required_request_body = request_body.get("required")
          required_body_params = []
          literal_body_params = []

          content = request_body.get("content", {})
          for mimetype, media_type in content.items():
            schema = media_type.get('schema')
            if schema:
              self.__extract_param_properties(components, required_body_params, {'tmp': schema}, literal_body_params)

            if literal_body_params:
              endpoint['literal_body_params'] = literal_body_params

            # Only support first media type
            break

          literal_query_params = endpoint.get('literal_query_params')
          if literal_query_params:
            self.__convert_literal_component_param(endpoint, required_query_params, literal_query_params, 'query_param_name', 'literal_query_params')
            
          literal_body_params = endpoint.get('literal_body_params')
          if literal_body_params:
            self.__convert_literal_component_param(endpoint, required_body_params, literal_body_params, 'body_param_name', 'literal_body_params')

          # Responses -> construct lists of response header and response param name resources
          responses = operation.get('responses', {})
          if responses:
            self.__parse_responses(endpoint, responses, components)

          endpoints.append(endpoint)
    
    return endpoints

  def __add_info(self, file_data: Dict):
    if 'info' in file_data:
      return

    file_data['info'] = {
      'version': '0.0.1',
      'title': ''
    }

  def __is_missing_oauth2_scopes(self, file_data: Dict):
    is_missing = False
    if 'components' in file_data:
      components = file_data['components']

      if 'securitySchemes' in components:
        security_schemes = components['securitySchemes']

        for scheme_name, scheme in security_schemes.items():
          if scheme.get('type') == 'oauth2':
            flows = scheme.get('flows')

            if flows and flows.get('clientCredentials'):
              scopes = flows.get('clientCredentials').get('scopes')

              if scopes is None:
                return True

            break

    return is_missing

  def __add_oauth2_default_scopes(self, file_data: Dict):
    if 'components' not in file_data:
      return
    components = file_data['components']

    if 'securitySchemes' not in components:
      return
    security_schemes = components['securitySchemes']

    oauth_security_scheme = {}
    for scheme_name, scheme in security_schemes.items():
      if scheme.get('type') == 'oauth2':
        oauth_security_scheme = scheme
        break

    # If empty dict or None
    if not oauth_security_scheme:
      return

    flows = oauth_security_scheme['flows']
    if not flows:
      return

    client_credentials = flows['clientCredentials']
    if not client_credentials:
      return

    client_credentials['scopes'] = {}

  def __get_most_recent_param(self, literal_params: dict):
    return list(literal_params)[-1] if literal_params else None

  def __get_second_most_recent_param(self, literal_params: dict):
    if not literal_params or len(literal_params) < 2:
      return None

    return list(literal_params)[-2]

  def __extract_param_properties(self, components, required_component_params, schema_object, literal_component_params, curr_id=0, parent_id=None, parent=None, query_string=''):
    # Name of the schema object (i.e. the name of the body_param_name or response_param_name component)
    property_name = list(schema_object.keys())[0]

    # The actual OpenAPI schema object
    property_schema = list(schema_object.values())[0] 

    if '$ref' in property_schema.keys():
      property_schema = self.__dereference(components, property_schema.get('$ref'))

    # Generates the entire query string for this particular property (ex: Pets[*].name for property 'name' of a 'Pet' element in the 'Pets' array
    if property_name == 'tmp':
      query = query_string 
    else:
      if property_name.endswith('Element'):
        query = f"{query_string}[*]"
      elif query_string == '':
        query = property_name
      else:
        query = f"{query_string}.{property_name}"

    # A schema is traversable if it has one more non-empty nested schema objects
    traversable = property_schema.get('properties') or property_schema.get('items') or property_schema.get('allOf')
    
    if not traversable:
      # Ex: {'name': {'type': 'string', 'description': 'Name of pet', 'Example': 'Buddy'}}
      if property_name != 'tmp':
        literal_val_type = self.__open_api_to_default_python_type(property_schema.get('type', 'object'))
        literal_val = {'name': property_name, 'value': literal_val_type, 'required': property_schema.get('required', False), 'query': query, 'id': curr_id, 'parent_id': parent_id}
        literal_component_params.append(literal_val)
      return curr_id

    if 'properties' in property_schema.keys():
      # Ex: {'tmp': {'type': 'object', 'required': ['name'], 'properties': {'name': {'type': 'string'}, 'tag': {'type': 'string'}}}}
      required_component_params += property_schema.get('required', [])

      # curr_id_tmp is the parent_id of all elements in param_properties
      # Should be set to parent_id when property_name is 'tmp' to handle case where a schema uses the 'allOf' keyword since curr_id will change between every recursive call on each 'part' element in the allOf list
      curr_id_tmp = parent_id if property_name == 'tmp' else curr_id 

      literal_val = parent if property_name == 'tmp' else {'name': property_name, 'value': {}, 'required': property_schema.get('schema_required', False), 'query': query, 'id': curr_id, 'parent_id': parent_id}

      if property_name != 'tmp':
        literal_component_params.append(literal_val)

      param_properties = property_schema.get('properties')
      for property_key, property_value in param_properties.items():
        if property_key in required_component_params:
          if property_value.get('type') == 'object':
            # Necessary to avoid overwriting the 'required' list of an object-type schema
            property_value['schema_required'] = True
          else:
            property_value['required'] = True
          required_component_params.remove(property_key)
        curr_id = self.__extract_param_properties(components, required_component_params, {property_key: property_value}, literal_component_params, curr_id=curr_id+1, parent_id=curr_id_tmp, parent=literal_val, query_string=query)    

    elif property_schema.get('type') == 'array':
      # Ex: {'tmp': {'type': 'array', 'items': {'$ref': '#/components/schemas/NewPet'}}}
      if property_name == 'tmp':
        schema_object = {"Element": property_schema['items']}
        curr_id = self.__extract_param_properties(components, required_component_params, schema_object, literal_component_params, curr_id=curr_id+1, parent_id=parent_id, parent=parent, query_string=query)
      else:
        schema_object = {f"{property_name}Element": property_schema['items']}
        literal_val = {'name': property_name, 'value': [], 'required': property_schema.get('required', False), 'query': query, 'id': curr_id, 'parent_id': parent_id}
        literal_component_params.append(literal_val)
        curr_id = self.__extract_param_properties(components, required_component_params, schema_object, literal_component_params, curr_id=curr_id+1, parent_id=curr_id, parent=parent, query_string=query)

    elif 'allOf' in property_schema.keys():
      # Ex: {'Element': {'type': 'object', 'allOf': [{'$ref': '#/components/schemas/NewPet'}, {'type': 'object', 'required': ['id'], 'properties': {'id': {'type': 'integer', 'format': 'int64'}}}]}}
      all_of = property_schema.get('allOf')
      literal_val = parent if property_name == 'tmp' else {'name': property_name, 'value': {}, 'required': property_schema.get('self_required', False), 'query': query, 'id': curr_id, 'parent_id': parent_id}
      if property_name != 'tmp':
        literal_component_params.append(literal_val)

      curr_id_tmp = parent_id if property_name == 'tmp' else curr_id 
      for part in all_of:
        required_component_params = part.get('required', [])
        curr_id = self.__extract_param_properties(components, required_component_params, {'tmp': part}, literal_component_params, curr_id=curr_id, parent_id=curr_id_tmp, parent=literal_val, query_string=query)

    return curr_id
  

  # Returns the schema object located at the given reference path
  def __dereference(self, components: SchemaPath, reference: str):
    # '#/components/schemas/NewPet'
    if not reference.startswith('#'):
      print('external references are not supported yet')
    if not reference.startswith('#/components'):
      print('non component references are not supported yet')
    else:
      ref_split = reference.split('#/components/')
      component_data = ref_split[1].split('/')
      component_type = component_data[0]
      component_name = component_data[1]
      component = components.get(component_type, {})

      # Example: {'type': 'object', 'required': ['name'], 'properties': {'name': {'type': 'string'}, 'tag': {'type': 'string'}}}
      body_spec = component.content()[component_name]
        
      if '$ref' in body_spec.keys():
        body_spec = self.__dereference(components, body_spec.get('$ref'))
      
      return body_spec 
  
  def __convert_literal_component_param(self, endpoint: EndpointShowResponse,
      required_component_params: List[str], literal_component_params: Union[dict, list],
      component_name: str, literal_component_name: str) -> None:

    if not literal_component_params:
      return

    builder = SchemaBuilder(endpoint['id'], component_name)
    built_params = builder.build(literal_component_params)

    built_params_list = list(built_params)
    for param in built_params_list:
      if param.get('is_required', False) == True:
        inferred_type = param['inferred_type']
        default_python_type = convert_reverse(inferred_type)

        if not param.get('values'):
          param['values'] = []
        param['values'].append(default_python_type)
      else:
        param['is_required'] = False

    if not endpoint.get(component_name + 's'):
      endpoint[component_name + 's'] = built_params_list
    else:
      endpoint[component_name + 's'].extend(built_params_list)

    del endpoint[literal_component_name]

  # urllib.parse.urljoin() doesn't work for some of our edge cases
  # and results in missing path components so use the custom __urljoin
  # https://stackoverflow.com/a/58037371
  def __join_slash(self, a: str, b: str):
    return a.rstrip('/') + '/' + b.lstrip('/')

  def __urljoin(self, *args: str):
    return reduce(self.__join_slash, args) if args else ''

  def __to_simple_dict(self, complex_dict):
    simple_dict = {}
    for key, val in complex_dict.items():
      simple_dict[key] = val['value']

    return simple_dict


  def __build_url(self, host, scheme, port, path, query):
    s = host
    if scheme and len(scheme) > 0:
      s = f"{scheme}://{s}"

    if port != None:
      if not ((scheme == 'https' and port == 443) or (scheme == 'http' and port == 80)):
        s = f"{s}:{port}"

    s += path

    _query = query
    if _query and len(_query) > 0:
      s = f"{s}?{_query}"

    return s

  # https://swagger.io/docs/specification/data-models/data-types/
  def __convert_open_api_type(self, open_api_type: str) -> str:
    type_map = {}
    type_map['string'] = str(str)
    type_map['number'] = str(float)
    type_map['integer'] = str(int) 
    type_map['boolean'] = str(bool) 
    type_map['array'] = str(list)
    type_map['object'] = str(dict)

    return type_map[open_api_type]

  def __open_api_to_default_python_type(self, open_api_type: str):
    type_map = {}
    type_map['string'] = str()
    type_map['number'] = float()
    type_map['integer'] = int()
    type_map['boolean'] = bool()
    type_map['array'] = list()
    type_map['object'] = dict()

    return type_map[open_api_type]

  def __num_variables(self, url: str) -> int:
    num = url.count('{')
    return num

  def __evaluate_servers(self, servers: SchemaPath) -> List[dict]:
    result = []

    if not servers:
      default_server = {'url': '/'}
      return [default_server]

    for server in servers:
      original_url = server["url"]

      variables = server.get("variables", {})
      if not variables:
        result.append({'url': original_url})
        continue

      split_url = re.split('{|}', original_url)
      existent_vars = {}
      for part in split_url:
        if part in variables.keys():
          existent_vars[part] = variables[part]

      var_to_possible_vals = {}
      all_possible_vals = []

      for variable_name, variable in existent_vars.items():
        enum_list = variable.get('enum')
        if enum_list:
          var_to_possible_vals[variable_name] = enum_list
        else:
          default_value = variable['default']
          var_to_possible_vals[variable_name] = [default_value]

      all_possible_vals = list(var_to_possible_vals.values())
      product = list(itertools.product(*all_possible_vals))

      for permutation in product:
        split_url_copy = copy.deepcopy(split_url)

        for i, part in enumerate(split_url_copy):
          for j, var in enumerate(var_to_possible_vals.keys()):
            if var == part:
              split_url_copy[i] = permutation[j]
              break

        evaluated_url = ''.join(split_url_copy)
        result.append({'url': evaluated_url})
    
    return result

  def __parse_responses(self, endpoint: EndpointShowResponse, responses: SchemaPath, components: SchemaPath):
    for response_code, response_definition in responses.items():
      # Only support status code 200 for now
      if response_code != '200':
        continue

      # Construct response param name components
      literal_response_params = []
      required_response_params = []
      response_content = response_definition.get('content', {})
      for mimetype, media_type in response_content.items():
        schema = media_type.get('schema')
        if schema:
          self.__extract_param_properties(components, required_response_params, {'tmp': schema}, literal_response_params)
        
        if literal_response_params:
          endpoint['literal_response_params'] = literal_response_params

        # Only support first media type
        break
      
      literal_response_params = endpoint.get('literal_response_params')
      if literal_response_params:
        self.__convert_literal_component_param(endpoint, required_response_params, literal_response_params, 'response_param_name', 'literal_response_params')

      # Construct response header name components
      response_headers = response_definition.get('headers', {})
      for header_name, header_definition in response_headers.items():
        response_header_name: RequestComponentName = {}
        response_header_name['name'] = header_name
        
        if '$ref' in header_definition:
          header_definition = self.__dereference(components, header_definition.get('$ref'))

        header_example = header_definition.get('example')
        if header_example:
          response_header_name['value']= header_example
        response_header_name['is_required'] = header_definition.get('is_required', False)
        response_header_name['is_deterministic'] = True
          
        if not endpoint.get('response_header_names'):
          endpoint['response_header_names'] = []
        endpoint['response_header_names'].append(response_header_name)	  



import pdb
import requests

from mitmproxy.http import Request as MitmproxyRequest

from stoobly_agent.app.models.adapters.python import PythonRequestAdapterFactory

class TestMitmproxyAdapter():

  def test_www_google_com(self):
    request = requests.Request(
      data=b'',
      headers={
        'Content-Length': '0'
      },
      method='GET',
      url='https://www.google.com',
    )

    mitmproxy_request = PythonRequestAdapterFactory(request).mitmproxy_request()
    self.__test_equivalence(mitmproxy_request, request)

  def test_stoobly_com(self):
    data = b'{"name": "test", "description": "test"}'
    request = requests.Request(
      data=data,
      headers={
        'Content-Length': str(len(data)),
        'Content-Type': 'application/json',
      },
      method='POST',
      url='https://stoobly.com/scenarios',
    )

    mitmproxy_request = PythonRequestAdapterFactory(request).mitmproxy_request()

    self.__test_equivalence(mitmproxy_request, request)

  def __test_equivalence(self, mitmproxy_request: MitmproxyRequest, request: requests.Request):
    assert mitmproxy_request.method == request.method
    assert mitmproxy_request.url.rstrip('/') == request.url
    assert mitmproxy_request.content == request.data

    for key, val in request.headers.items():
      assert val == mitmproxy_request.headers.get(key)



from pathlib import Path
from typing import Dict

import pytest

from stoobly_agent.app.cli.helpers.openapi_endpoint_adapter import (
  OpenApiEndpointAdapter,
)


@pytest.mark.openapi
class TestOpenApiEndpointAdapterAdditionalProps():
  @pytest.fixture(scope='class')
  def mock_data_directory_path(self):
    return Path(__file__).parent.parent.parent.parent / 'mock_data'

  @pytest.fixture(scope='class')
  def petstore_file_path(self, mock_data_directory_path):
    path = mock_data_directory_path / "petstore-additional-props.yaml"
    return path

  @pytest.fixture(scope='class')
  def open_api_endpoint_adapter(self):
    adapter = OpenApiEndpointAdapter()
    return adapter

  @pytest.fixture(scope='class')
  def expected_post_pets_endpoint(self) -> Dict:
    return {
      "id": 1,
      "method": "POST",
      "host": "petstore.swagger.io",
      "match_pattern": "/v1/pets",
      "path": "/v1/pets",
      "port": "80",
      "body_param_names": [
        {
          "endpoint_id": 1,
          "inferred_type": "String",
          "is_required": False,
          "is_deterministic": True,
          "name": "name",
          "query": "name",
          "body_param_name_id": None,
          "id": 1
        }
      ],
      'path_segment_names': [
        {
          'name': 'v1',
          'type': 'static'
        },
        {
          'name': 'pets',
          'type': 'static'
        }
      ],
    }


  def test_adapt_from_file(self, open_api_endpoint_adapter, petstore_file_path, expected_post_pets_endpoint):
    adapter = open_api_endpoint_adapter

    endpoints = adapter.adapt_from_file(petstore_file_path)

    assert len(endpoints) == 1
    post_pets_endpoint = endpoints[0]
    assert post_pets_endpoint == expected_post_pets_endpoint




from .resource_key import ResourceKey

LOCAL_ORGANIZATION_ID = 0

class InvalidOrganizationKey(Exception):
  pass

class OrganizationKey(ResourceKey):

  def __init__(self, key: str):
    super().__init__(key)

    if not self.id:
      raise InvalidOrganizationKey('Missing id')

  @property
  def id(self) -> str:
    return self.get('i')

  @staticmethod
  def encode(organization_id: str):
    return ResourceKey.encode({
      'i': organization_id,
    })


from stoobly_agent.app.proxy.context import InterceptContext
from stoobly_agent.app.proxy.mock.context import MockContext
from stoobly_agent.app.proxy.record.context import RecordContext
from stoobly_agent.app.proxy.replay.context import ReplayContext
from stoobly_agent.app.proxy.test.context import TestContext

def handle_before_request(context: InterceptContext):
  print('before_request')

def handle_before_record(context: RecordContext):
  print('before_record')

def handle_before_mock(context: MockContext):
  print('before_mock')

def handle_before_replay(context: ReplayContext):
  print('before_replay')

def handle_before_test(context: ReplayContext):
  print('before_test')

def handle_after_record(context: RecordContext):
  print('after_record')

def handle_after_mock(context: MockContext):
  print('after_mock')

def handle_after_replay(context: ReplayContext):
  print('after_replay')

def handle_after_test(context: TestContext):
  print('after_test')

def handle_before_response(context: InterceptContext):
  print('before_response')


import os

from typing import Literal

from stoobly_agent.config.data_dir import CERTS_DIR_NAME, DATA_DIR_NAME

APP_DIR = '${APP_DIR}'
APP_DIR_ENV = 'APP_DIR'
APP_NETWORK_ENV = 'APP_NETWORK'
APP_NAME_ENV = 'APP_NAME'
BIN_FOLDER_NAME = 'bin'
CA_CERTS_DIR_ENV = 'CA_CERTS_DIR'
CERTS_DIR_ENV = 'CERTS_DIR'
COMPOSE_TEMPLATE = '.docker-compose.{workflow}.yml'
CONFIG_FILE = '.config.yml'
CONTEXT_DIR_ENV = 'CONTEXT_DIR'
DOCKER_NAMESPACE = 'docker'
ENV_FILE = '.env'
NAMESERVERS_FILE = '.nameservers'
PUBLIC_FOLDER_NAME = 'public'
SERVICE_DETACHED = '${SERVICE_DETACHED}'
SERVICE_DETACHED_ENV = 'SERVICE_DETACHED'
SERVICE_HOSTNAME = '${SERVICE_HOSTNAME}'
SERVICE_HOSTNAME_ENV = 'SERVICE_HOSTNAME'
SERVICE_DNS = '${SERVICE_DNS}'
SERVICE_DNS_ENV = 'SERVICE_DNS'
SERVICE_NAME = '${SERVICE_NAME}'
SERVICE_NAME_ENV = 'SERVICE_NAME'
SERVICE_PROXY_MODE = '${SERVICE_PROXY_MODE}'
SERVICE_PROXY_MODE_ENV = 'SERVICE_PROXY_MODE'
SERVICE_SCHEME = '${SERVICE_SCHEME}'
SERVICE_SCHEME_ENV = 'SERVICE_SCHEME'
SERVICE_PORT = '${SERVICE_PORT}'
SERVICE_PORT_ENV = 'SERVICE_PORT'
SERVICE_PRIORITY_ENV = 'SERVICE_PRIORITY'
SERVICE_SCRIPTS = '${SERVICE_SCRIPTS}'
SERVICE_SCRIPTS_DIR = '/usr/local/bin/services'
SERVICE_SCRIPTS_ENV = 'SERVICE_SCRIPTS'
STOOBLY_HOME_DIR = '/home/stoobly'
STOOBLY_DATA_DIR = os.path.join(STOOBLY_HOME_DIR, DATA_DIR_NAME)
STOOBLY_CERTS_DIR = os.path.join(STOOBLY_DATA_DIR, CERTS_DIR_NAME)
USER_ID_ENV = 'USER_ID'
WORKFLOW_CONTAINER_CONFIGURE = 'configure'
WORKFLOW_CONTAINER_INIT = 'init'
WORKFLOW_CONTAINER_PROXY = 'proxy'
WORKFLOW_CONTAINER_TEMPLATE = '{service_name}.{container}'
WORKFLOW_CONTAINER_CONFIGURE_TEMPLATE = '{service_name}.' + WORKFLOW_CONTAINER_CONFIGURE
WORKFLOW_CONTAINER_INIT_TEMPLATE = '{service_name}.' + WORKFLOW_CONTAINER_INIT
WORKFLOW_CONTAINER_PROXY_TEMPLATE = '{service_name}.' + WORKFLOW_CONTAINER_PROXY
WORKFLOW_EXEC_TYPE = 'exec'
WORKFLOW_MOCK_TYPE = 'mock'
WORKFLOW_NAME = '${WORKFLOW_NAME}'
WORKFLOW_NAME_ENV = 'WORKFLOW_NAME'
WORKFLOW_NAMESPACE_ENV = 'WORKFLOW_NAMESPACE'
WORKFLOW_RECORD_TYPE = 'record'
WORKFLOW_SCRIPTS = '${WORKFLOW_SCRIPTS}'
WORKFLOW_SCRIPTS_DIR = '/usr/local/bin/workflows'
WORKFLOW_SCRIPTS_ENV = 'WORKFLOW_SCRIPTS'
WORKFLOW_TEMPLATE = '${WORKFLOW_TEMPLATE}'
WORKFLOW_TEMPLATE_ENV = 'WORKFLOW_TEMPLATE'
WORKFLOW_TEST_TYPE = 'test'

WORKFLOW_TEMPLATE_OPTION = Literal[WORKFLOW_MOCK_TYPE, WORKFLOW_RECORD_TYPE, WORKFLOW_TEST_TYPE]



import pdb

from stoobly_agent.app.models.adapters.python.request.mitmproxy_adapter import MitmproxyRequestAdapter as PythonRequestMitmproxyRequestAdapter
from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.lib.orm.request import Request

class MitmproxyRequestAdapter():

  def __init__(self, request: Request):
    self.__request = request

  def adapt(self):
    raw_http_request_adapter = RawHttpRequestAdapter(self.__request.raw)
    python_request = raw_http_request_adapter.to_request()
    protocol = raw_http_request_adapter.protocol
    http_version = protocol.split('/')[1]

    return PythonRequestMitmproxyRequestAdapter(http_version, python_request).adapt()



from stoobly_agent.config.constants import mock_policy, record_policy, replay_policy, test_strategy

from .types.proxy_settings import DataRules as IDataRules

class DataRules:

  def __init__(self, data_rules: IDataRules):
    self.__data_rules = data_rules or {}

    self.__mock_policy = self.__data_rules.get('mock_policy') or mock_policy.FOUND
    self.__record_policy = self.__data_rules.get('record_policy') or record_policy.ALL
    self.__replay_policy = self.__data_rules.get('replay_policy') or replay_policy.ALL
    self.__scenario_key = self.__data_rules.get('scenario_key')
    self.__test_policy = self.__data_rules.get('test_policy') or mock_policy.FOUND
    self.__test_strategy = self.__data_rules.get('test_strategy') or test_strategy.DIFF

  @property
  def mock_policy(self):
    return self.__mock_policy 

  @mock_policy.setter
  def mock_policy(self, v):
    self.__mock_policy = v
    self.__data_rules['mock_policy'] = v

  @property
  def record_policy(self):
    return self.__record_policy 

  @record_policy.setter
  def record_policy(self, v):
    self.__record_policy = v
    self.__data_rules['record_policy'] = v

  @property
  def replay_policy(self):
    return self.__replay_policy

  @replay_policy.setter
  def replay_policy(self, v):
    self.__record_policy = v
    self.__data_rules['replay_policy'] = v

  @property
  def scenario_key(self):
    return self.__scenario_key

  @scenario_key.setter
  def scenario_key(self, v):
    self.__scenario_key = v
    self.__data_rules['scenario_key'] = v
    if not v and 'scenario_key' in self.__data_rules:
      del self.__data_rules['scenario_key']

  @property
  def test_policy(self):
    return self.__test_policy 

  @test_policy.setter
  def test_policy(self, v):
    self.__test_policy = v
    self.__data_rules['test_policy'] = v

  @property
  def test_strategy(self):
    return self.__test_strategy 

  def to_dict(self) -> IDataRules:
    return {
      'mock_policy': self.__mock_policy,
      'record_policy': self.__record_policy,
      'replay_policy': self.__replay_policy,
      'scenario_key': self.__scenario_key,
      'test_policy': self.__test_policy,
      'test_strategy': self.__test_strategy,
    }


import base64
import json

class ResourceKey():

    def __init__(self, key):
        self.__raw = key

        try:
            key = base64.b64decode(key)
        except:
            self.decoded_key = {}

            return

        # TODO: add specific error catching
        try:
            self.decoded_key: dict = json.loads(key)
        except:
            self.decoded_key = {}

    @property
    def raw(self):
        return self.__raw

    @staticmethod
    def encode(data: dict):
        return base64.b64encode(json.dumps({
            **data,
        }).encode())

    def get(self, k: str):
        v = self.decoded_key.get(k)
        return str(v) if v != None else None


from typing import Literal, TypedDict

REQUEST_RESOURCE = 'request'
SCENARIO_RESOURCE = 'scenario'

DELETE_ACTION = 'DELETE'
PUT_ACTION = 'PUT'

Action = Literal[f"{DELETE_ACTION}", f"{PUT_ACTION}"]
Resource = Literal[f"{REQUEST_RESOURCE}", f"{SCENARIO_RESOURCE}"]

class SnapshotOptions(TypedDict):
  action: Action

class RequestSnapshotOptions(SnapshotOptions):
  decode: bool


import pdb

from typing import List

from stoobly_agent.app.cli.helpers.trace_aliases import parse_aliases
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.api.endpoints_resource import EndpointsResource
from stoobly_agent.lib.orm.trace import Trace

class TraceContextFacade():

  def __init__(self, settings: Settings, trace: Trace = None):
    self.__settings = settings
    self.__endpoints_resource = EndpointsResource(self.__settings.remote.api_url, self.__settings.remote.api_key)
    self.__trace_context = TraceContext(self.__endpoints_resource, trace)

  @property
  def trace_context(self):
    return self.__trace_context

  def with_aliases(self, aliases):
    for _alias in self.__parse_aliases(aliases):
      self.__trace_context.create_trace_alias(_alias['name'], _alias['value'])
    return self

  def __parse_aliases(self, aliases: List[str]):
    if not aliases:
      return []

    return parse_aliases(aliases) 


import pdb
import pytest
import time

from typing import List

from stoobly_agent.app.models.factories.resource.local_db.helpers.tiebreak_scenario_request import (
  access_request, generate_session_id, reset_sessions, tiebreak_scenario_request
)
from stoobly_agent.lib.cache import Cache
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.test.test_helper import reset

class RequestMock():

  def __init__(self, id):
    self.__id = id

  @property
  def id(self):
    return self.__id

@pytest.fixture(autouse=True, scope='class')
def settings():
  return reset()

class TestTiebreakScenarioRequest():
  @pytest.fixture(scope='class')
  def created_request_one(self):
    return RequestMock(1)

  @pytest.fixture(scope='class')
  def created_request_two(self):
    return RequestMock(2)

  @pytest.fixture(scope='class')
  def created_request_three(self):
    return RequestMock(3)

  @pytest.fixture(scope='class')
  def session_id(self):
    return generate_session_id({
      'scenario_id': 'test'
    })

  @pytest.fixture(scope='class')
  def requests(self, created_request_one: Request, created_request_two: Request):
    return [created_request_one, created_request_two]

  def test_it_returns_request_one(self, session_id: str, requests: List[Request]):
    request = tiebreak_scenario_request(session_id, requests)
    assert request.id == 1

  def test_it_returns_request_two(self, session_id: str, requests: List[Request]):
    access_request(session_id, requests[0].id)
    request = tiebreak_scenario_request(session_id, requests)
    assert request.id == 2

  def test_it_returns_request_two_again(self, session_id: str, requests: List[Request]):
    request = tiebreak_scenario_request(session_id, requests)
    assert request.id == 2

  def test_it_returns_request_three(self, session_id: str, requests: List[Request]):
    access_request(session_id, requests[0].id)
    request = tiebreak_scenario_request(session_id, requests)
    assert request.id == 2

  def test_it_returns_request_one_again(self, session_id: str, requests: List[Request]):
    access_request(session_id, requests[1].id)
    request = tiebreak_scenario_request(session_id, requests)
    assert request.id == 1

  class TestWhenTimeout():
    @pytest.fixture(autouse=True, scope='class')
    def cache(self):
      cache = Cache.instance()
      return cache

    def test_it_returns_request_one(self, session_id: str, requests: List[Request]):
      access_request(session_id, requests[0].id, 100)

      time.sleep(1)

      request = tiebreak_scenario_request(session_id, requests)
      assert request.id == 1

    def test_it_returns_request_two(self, session_id: str, requests: List[Request]):
      access_request(session_id, requests[0].id, 100)

      time.sleep(0.05)

      request = tiebreak_scenario_request(session_id, requests)
      assert request.id == 2

  class TestWhenResetSessions():
    @pytest.fixture(scope='class')
    def created_request_one(self):
      return RequestMock(1)

    @pytest.fixture(autouse=True, scope='class')
    def cache(self):
      cache = Cache.instance()
      return cache

    def test_it_resets(self, cache: Cache, created_request_one: Request):
      cache.write('persists', 1)
      access_request('1', created_request_one.id)
      reset_sessions()

      assert cache.read('persists') != None
      assert len(cache.read_all()) == 1



import pdb

from typing import Callable, TypedDict

from stoobly_agent.app.settings import Settings
from stoobly_agent.lib.utils.conditional_decorator import ConditionalDecorator

class LifeCycleOptions(TypedDict):
  before_action: Callable[[], None]
  after_action: Callable[[], None]

class CommandOptions(TypedDict):
  disable: LifeCycleOptions
  enable: LifeCycleOptions
  show: LifeCycleOptions

class ConfigDecorator():
  __main = None
  __settings = None
  __setting = None # e.g. features.remote

  def __init__(self, main, settings: Settings, setting: str, options: CommandOptions = {}):
    self.__main = main
    self.__options = options
    self.__settings = settings
    self.__setting = setting

  def decorate(self):
    self.__with_enable()
    self.__with_disable()
    self.__with_show()
    return self.__main

  def __with_enable(self):
    life_cycle_options = self.__options.get('enable') or {}

    @self.__main.command()
    @ConditionalDecorator(life_cycle_options.get('set_options'), bool(life_cycle_options.get('set_options')))
    def enable(**kwargs):
      if life_cycle_options:
        before_action = life_cycle_options['before_action']
        if isinstance(before_action, Callable):
          before_action(**kwargs)

      settings_hash = self.__settings.to_dict()
      self.__resolve(settings_hash, self.__setting, value=True)
      self.__settings.write(settings_hash)
      print(f"{self.__setting} enabled")

  def __with_disable(self):
    @self.__main.command()
    def disable():
      settings_hash = self.__settings.to_dict()
      self.__resolve(settings_hash, self.__setting, value=False)
      self.__settings.write(settings_hash)
      print(f"{self.__setting} disabled")

  def __with_show(self):
    @self.__main.command()
    def show():
      settings_hash = self.__settings.to_dict()
      print(self.__resolve(settings_hash, self.__setting))

  def __resolve(self, settings_hash, setting: str, **kwargs):
    keys = setting.split('.')
    length = len(keys)

    i = 0
    t = settings_hash
    for key in keys:
      if type(t) == dict:
        if i == length - 1 and 'value' in kwargs:
          t[key] = kwargs['value']

        t = t.get(key) 
      elif type(t) == list:
        if i == length - 1 and 'value' in kwargs:
          t[key] = kwargs['value']

        t = t[key]
      else:
        return None

      if not t:
        break

      i += 1

    return t



import pdb
import requests

from stoobly_agent.app.proxy.mitmproxy.response_facade import MitmproxyResponseFacade
from stoobly_agent.app.proxy.record.response_string import ResponseString

from .mitmproxy_adapter import MitmproxyResponseAdapter

class RawResponseAdapter():

  def __init__(self, response: requests.Response):
    self.__response = response

  def adapt(self):
    mitmproxy_response = MitmproxyResponseAdapter(self.__response).adapt()
    adapted_response = MitmproxyResponseFacade(mitmproxy_response)
    response_string = ResponseString(adapted_response, None) 
    return response_string.get()


from io import BytesIO
from requests import Response

class RequestsResponseBuilder():

  def __init__(self):
    self.__response = Response()

  def with_header(self, name, value):
    self.__response.headers[name] = value
    return self

  def with_body(self, content: bytes):
    self.__response.raw = BytesIO(content)
    return self

  def with_status_code(self, code: int):
    self.__response.status_code = code
    return self

  def with_reason(self, reason: str):
    self.__response.reason = reason
    return self

  def with_content_encoding(self, encoding):
    self.__response.encoding = encoding
    return self

  def build(self):
    return self.__response


import errno
import os
import pdb
import time
import tempfile

from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow
from mitmproxy.http import Request as MitmproxyRequest

from stoobly_agent.app.models.request_model import RequestModel
from stoobly_agent.app.proxy.intercept_settings import InterceptSettings
from stoobly_agent.app.proxy.record import JoinedRequest, RequestString, ResponseString
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import request_origin
from stoobly_agent.lib.api.param_builder import ParamBuilder
from stoobly_agent.lib.logger import Logger, bcolors
from stoobly_agent.lib.orm.request import Request

from ..utils.publish_change_service import publish_requests_modified
from .join_request_service import join_request_from_flow

AGENT_STATUSES = {
    'REQUESTS_MODIFIED': 'requests-modified'
}

LOG_ID = 'Record'
NAMESPACE_FOLDER = 'stoobly'

def inject_upload_request(request_model: RequestModel, intercept_settings: InterceptSettings):
    settings = Settings.instance()

    if not request_model:
        request_model = RequestModel(settings)

    if not intercept_settings:
        intercept_settings = InterceptSettings(settings)

    return lambda flow: upload_request(request_model, intercept_settings, flow)

###
#
# Upon receiving a response, create the request in API for future use
#
# @param request_model [StooblyApi]
# @param settings [Settings.mode.mock | Settings.mode.record]
# @param res [Net::HTTP::Response]
#
def upload_request(
    request_model: RequestModel, intercept_settings: InterceptSettings, flow: MitmproxyHTTPFlow = None
):
    Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Recording{bcolors.ENDC} {flow.request.url}")

    joined_request = join_request_from_flow(flow, intercept_settings)

    project_key = intercept_settings.project_key
    scenario_key = intercept_settings.scenario_key
    body_params = __build_body_params(
        project_key,
        joined_request, 
        flow=flow,
        scenario_key=scenario_key
    )

    # If request_origin is WEB, then we are in proxy
    # This means that we have access to Cache singleton and do not need send a request to update the status
    sync = intercept_settings.request_origin == request_origin.WEB 
    res = __upload_request_with_body_params(request_model, body_params, sync)
 
    if intercept_settings.settings.is_debug():
        file_path = __debug_request(flow.request, joined_request.build())
        Logger.instance(LOG_ID).debug(f"Writing request to {file_path}")
    elif not res:
        file_path = __debug_request(flow.request, joined_request.build())
        Logger.instance(LOG_ID).error(f"Error: Failed to upload, writing request to {file_path}")

    return res

def upload_staged_request(
    request: Request, request_model: RequestModel, project_key: str, scenario_key: str = None
):
    Logger.instance(LOG_ID).info(f"{bcolors.OKCYAN}Recording{bcolors.ENDC} {request.url}")

    response = request.response

    if not response:
        return

    request_string = RequestString(None)
    request_string.set(request.raw)
    request_string.control = request.control

    response_string = ResponseString(None, None)
    response_string.set(response.raw)
    response_string.control = response.control

    joined_request = JoinedRequest()
    joined_request.request_string = request_string
    joined_request.response_string = response_string

    body_params =__build_body_params(
        project_key,
        joined_request,
        scenario_key=scenario_key
    )

    return __upload_request_with_body_params(request_model, body_params)

def __upload_request_with_body_params(request_model: RequestModel, body_params: dict, sync=True):
    request, status = request_model.create(**body_params)

    if status < 400:
        publish_requests_modified(body_params['project_id'], sync=sync)

        return request

def __build_body_params(project_key: str, joined_request: JoinedRequest, **kwargs):
    flow: MitmproxyHTTPFlow = kwargs.get('flow')
    scenario_key = kwargs.get('scenario_key')

    body_params = ParamBuilder({ 'flow': flow, 'joined_request': joined_request })
    body_params.with_resource_scoping(project_key, scenario_key)

    return body_params.build()

# Write the request to a file to help debug
def __debug_request(request: MitmproxyRequest, raw_requests: bytes):
    # Build file path, replace slashes with underscores
    request_path = request.path.replace('/', '_')
    if len(request_path) > os.pathconf('/', 'PC_NAME_MAX'):
        request_path = f"{request_path[0:252]}..."

    timestamp = str(int(time.time() * 1000))
    file_path = os.path.join(tempfile.gettempdir(), NAMESPACE_FOLDER, request_path, timestamp)

    if not os.path.exists(os.path.dirname(file_path)):
        try:
            os.makedirs(os.path.dirname(file_path))
        except OSError as err: # Guard against race condition
            if err.errno != errno.EEXIST:
                raise err

    with open(file_path, 'wb') as f:
        f.write(raw_requests)

    return file_path


from stoobly_orator.migrations import Migration

class CreateRequests(Migration):

    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create('requests') as table:
            table.increments('id')
            table.string('scheme').nullable()
            table.string('method')
            table.text('host')
            table.text('path')
            table.integer('port')
            table.string('headers_hash').index()
            table.string('body_text_hash').index()
            table.string('query_params_hash').index()
            table.string('body_params_hash').index()
            table.binary('control')
            table.binary('raw')
            table.timestamp('pushed_at').nullable()
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop('requests')



import os
import shutil

from stoobly_agent.app.models.adapters.orm import JoinedRequestStringAdapter
from stoobly_agent.lib.orm.request import Request

from .snapshot import Snapshot
from .snapshot_types import RequestSnapshotOptions

class RequestSnapshot(Snapshot):

  def __init__(self, uuid: str):
    super().__init__(uuid)
    
    self.__requests_dir_path = self.data_dir.snapshots_requests_dir_path
  
    self.__backup = None

  @property
  def backup(self):
    return self.__backup

  @property
  def exists(self):
    return os.path.exists(self.path)

  @property
  def request(self):
    request_file_path = self.path

    if not os.path.exists(request_file_path):
      return

    with open(request_file_path, 'rb') as fp:
      return fp.read()

  @property
  def path(self):
    dir_path = os.path.join(self.__requests_dir_path, self.uuid[0:2])

    if not os.path.exists(dir_path):
      os.mkdir(dir_path)

    return os.path.join(dir_path, self.uuid)

  def backup(self):
    if os.path.exists(self.path):
      with open(self.path, 'rb') as fp:
        self.__backup = fp.read()

  def copy(self, dest_dir: str):
    request_file_path = self.path
    return self.copy_file(request_file_path, dest_dir)

  def find_resource(self):
    return Request.find_by(uuid=self.uuid)

  def remove(self):
    request_file_path = self.path

    if os.path.exists(request_file_path):
      os.remove(request_file_path)

  def rollback(self):
    with open(self.path, 'wb') as fp:
      fp.write(self.__backup)

  def write(self, request: Request, **options: RequestSnapshotOptions):
    adapter = JoinedRequestStringAdapter(request)

    if options.get('decode'):
      adapter.decode_response()

    text = adapter.adapt()
    self.write_raw(text)

  def write_raw(self, text):
    with open(self.path, 'wb') as fp:
      fp.write(text)



import pdb

from typing import Callable, List, TypedDict

from stoobly_agent.app.cli.helpers.trace_context_facade import TraceContextFacade
from stoobly_agent.app.proxy.replay.trace_context import TraceContext
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import alias_resolve_strategy, request_origin, test_filter, test_strategy
from stoobly_agent.lib.api.keys.project_key import ProjectKey
from stoobly_agent.lib.orm.trace import Trace

class ReplayCliOptions(TypedDict):
  alias_resolve_strategy: alias_resolve_strategy.AliasResolveStrategy
  assign: List[str]
  group_by: str
  host: str
  lifecycle_hooks_path: str
  on_response: Callable
  project_key: str
  public_directory_path: str
  record: bool
  response_fixtures_path: str
  scenario_key: str
  scheme: str
  trace: Trace

class TestCliOptions(ReplayCliOptions):
  filter: test_filter.TestFilter
  report_key: str
  save: bool
  strategy: test_strategy.TestStrategy

class ReplayFacade():

  def __init__(self, settings: Settings):
    self.__settings = settings

  def common_cli_options(self, cli_options: ReplayCliOptions) -> ReplayCliOptions:
    assign = cli_options.get('assign')
    trace_context = None
    trace_id = cli_options.get('trace_id')

    # If a trace_id is given, use it to find a trace
    trace = Trace.find_by(id=trace_id) if trace_id else None 
    if trace or assign:
      facade = TraceContextFacade(self.__settings, trace)

      # If assign is given, create default TraceAliases for the trace 
      if assign:
        facade.with_aliases(assign)

      trace_context = facade.trace_context

    return {
      'after_replay': cli_options.get('after_replay'),
      'alias_resolve_strategy': cli_options.get('alias_resolve_strategy'),
      'before_replay': cli_options.get('before_replay'),
      'group_by': cli_options.get('group_by'),
      'host': cli_options.get('host'),
      'lifecycle_hooks_path': cli_options.get('lifecycle_hooks_path'),
      'overwrite': cli_options.get('overwrite'),
      'public_directory_path': cli_options.get('public_directory_path'),
      'request_origin': request_origin.CLI,
      'response_fixtures_path': cli_options.get('response_fixtures_path'),
      'save': cli_options.get('save'),
      'scheme': cli_options.get('scheme'),
      'trace_context': trace_context,
    }

  def common_replay_cli_options(self, cli_options: ReplayCliOptions) -> ReplayCliOptions:
    common_cli_options = self.common_cli_options(cli_options)

    # Scenario key has no meaning unless mode is record
    # If mode is record, then setting scenario key will record requests into the specified scenario 
    if cli_options.get('record'):
      scenario_key = cli_options.get('scenario_key')
      if not scenario_key:
        data_rules = self.data_rules()
        common_cli_options['scenario_key'] = data_rules.scenario_key
      else:
        common_cli_options['scenario_key'] = scenario_key
    
    return common_cli_options

  def common_test_cli_options(self, cli_options: TestCliOptions) -> TestCliOptions:
    filter = cli_options.get('filter')
    if not filter:
      # TODO: add test_filter to data_rules
      pass

    strategy = cli_options.get('strategy')
    if not strategy:
        data_rule = self.data_rules()
        strategy = data_rule.test_strategy

    save = cli_options.get('save')

    return {
      'test_filter': filter or test_filter.ALL,
      'test_save_results': save,
      'test_strategy': strategy or test_strategy.DIFF,
      **self.common_cli_options(cli_options)
    }

  def data_rules(self):
    project_key = ProjectKey(self.__settings.proxy.intercept.project_key)
    return self.__settings.proxy.data.data_rules(project_key.id)


import pdb
import requests

from stoobly_agent.app.settings import Settings

from .factories.resource.replayed_response import ReplayedResponseResourceFactory
from .model import Model
from .types import ReplayeResponseCreateParams

class ReplayedResponseModel(Model):

  def __init__(self, settings: Settings, **options):
    super().__init__(settings, **options)

  def as_local(self):
    self.adapter = ReplayedResponseResourceFactory(self.settings.remote).local_db()
    self.is_local = True

  def as_remote(self):
    self.is_local = False

  def create(self, **body_params: ReplayeResponseCreateParams):
    try:
      return self.adapter.create(**body_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def index(self, **query_params):
    try:
      return self.adapter.index(**query_params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def mock(self, replayed_response_id: str):
    try:
      return self.adapter.mock(replayed_response_id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def activate(self, replayed_response_id: str):
    try:
      return self.adapter.activate(replayed_response_id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def raw(self, replayed_response_id: str):
    try:
      return self.adapter.raw(replayed_response_id)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)

  def update(self, replayed_response_id: str, **params) -> requests.Response:
    try:
      return self.adapter.update(replayed_response_id, **params)
    except requests.exceptions.RequestException as e:
      return self.handle_request_error(e)


import os
import yaml

from .constants import CONFIG_FILE

class Config():

  def __init__(self, dir: str, file_name = None):
    self.__dir = dir
    self.__path = os.path.join(dir, file_name or CONFIG_FILE)

  @property
  def dir(self):
    return self.__dir

  @property
  def path(self):
    return self.__path

  def read(self, source: str = None):
    if not source:
      source = self.path

    config = {}

    if os.path.exists(source):
      with open(source, 'r') as fp:
        config = yaml.safe_load(fp)
        
    return config

  def write(self, config: dict):
    config_path = os.path.join(self.path)
    with open(config_path, 'w') as fp:
      yaml.dump(config, fp)


import os
import pdb
import threading

from http.server import HTTPServer
from urllib.parse import urlparse

from stoobly_agent.app.settings import Settings
from stoobly_agent.config.constants import env_vars
from stoobly_agent.lib.logger import Logger

from .application_http_request_handler import ApplicationHTTPRequestHandler

LOG_ID = 'UI'

def start_server(host, port):
    httpd = HTTPServer((host, port), ApplicationHTTPRequestHandler)
    httpd.serve_forever()

def initialize(**kwargs):
    url = f"http://{kwargs['ui_host']}:{kwargs['ui_port']}"

    os.environ[env_vars.AGENT_URL] = url

    # TODO: debug why the following is needed or else logger won't print
    settings = Settings.instance()
    settings.ui.url = url
    settings.commit()

    return url

def run(url):
    parsed_url = urlparse(url)

    Logger.instance(LOG_ID).info(f"Server listening at {url}\n")

    thread = threading.Thread(target=start_server, args=(parsed_url.hostname, parsed_url.port))
    thread.start()



import pdb

from .python_adapter import PythonResponseAdapter

class RawJoinedResponseAdapterFactory():

  def __init__(self, joined_request: bytes):
    self.__request = joined_request

  def python_response(self):
    return PythonResponseAdapter(self.__request).adapt()


import pdb
import requests

from httptools import HttpResponseParser
from io import BytesIO
from mitmproxy.net import encoding
from requests.structures import CaseInsensitiveDict
from typing import Union
from urllib3 import HTTPResponse

from stoobly_agent.app.proxy.mitmproxy.response import Response
from stoobly_agent.lib.utils.decode import decode
from stoobly_agent.lib.orm.utils.response_parse_handler import Response as ResponseDict, ResponseParseHandler

CRLF = b'\r\n'
DEFAULT_HTTP_VERSION = b'HTTP/1.1'

HEADER_SEPARATOR = ', '

class RawHttpResponseAdapter():

  def __init__(self, req_text: bytes):
    self.__req_text = req_text
    req_lines = self.__req_text.split(CRLF)

    self.__parse_response_line(req_lines[0])

    ind = 1
    self.headers = CaseInsensitiveDict()
    while ind < len(req_lines) and len(req_lines[ind]) > 0:
        colon_ind = req_lines[ind].find(b':')
        header_key = req_lines[ind][:colon_ind]
        header_value = req_lines[ind][colon_ind + 1:]

        decoded_key = decode(header_key)
        decoded_value = decode(header_value).strip()
        if decoded_key in self.headers:
          self.headers[decoded_key] = HEADER_SEPARATOR.join([self.headers[decoded_key], decoded_value])
        else:
          self.headers[decoded_key] = decoded_value

        ind += 1

    ind += 1

    data_lines = req_lines[ind:] if ind < len(req_lines) else None
    self.body = CRLF.join(data_lines)

  def to_response(self):
    response = requests.Response()
    response.status_code = self.status
    response.headers = self.headers

    # Enforce proper Content-Length header
    response.headers['Content-Length'] = str(len(self.body))

    response.raw = HTTPResponse(
      body=BytesIO(self.body),
      decode_content=False,
      headers=self.headers,
      preload_content=False
    )

    #response.raw = self.__decode_body(self.body, content_encoding)

    return response

  def to_response_from_http_response_parser(self):
    response_dict: ResponseDict = {}
    parser = self.__new_parser(response_dict)
    response_dict['status_code'] = parser.get_status_code()

    requests_response = Response()
    requests_response.headers = CaseInsensitiveDict()
    for key, val in response_dict['headers'].items():
      _key = decode(key) 
      _val = decode(key)
      requests_response.headers[_key] = _val

    # Unless we can create an object with the stream method, have to self decode
    content_encoding = requests_response.headers.get('content-encoding')
    requests_response.raw = self.__decode_body(response_dict.get('body'), content_encoding)

    requests_response.reason = response_dict.get('status')
    requests_response.status_code = response_dict['status_code']

    # TODO: parse cookies, url. Not high priority, these aren't used
    
    return requests_response

  def __parse_response_line(self, response_line):
    response_parts = response_line.split(b' ')
    self.protocol = decode(response_parts[0]) if len(response_parts) > 2 else DEFAULT_HTTP_VERSION
    self.status = int(decode(response_parts[1]))
    self.reason = decode(response_parts[2]) if len(response_parts) > 2 else ''

  def __str__(self):
    headers = CRLF.join(f'{key}: {self.headers[key]}' for key in self.headers)
    return f'{self.protocol} {self.status} {self.reason}{CRLF}' \
            f'{headers}{CRLF}{CRLF}{self.body}'

  def __new_parser(self, response_dict):
    handler = ResponseParseHandler(response_dict)
    parser = HttpResponseParser(handler)
    parser.feed_data(memoryview(self.__req_text))
    return parser

  def __decode_body(self, body: bytes, content_encoding):
    if content_encoding:
      decoded_response = encoding.decode(body, content_encoding)
    else:
      decoded_response = body 

    return BytesIO(decoded_response)


from mitmproxy.http import HTTPFlow as MitmproxyHTTPFlow

from stoobly_agent.app.proxy.intercept_settings import InterceptSettings

from ..context import InterceptContext

class RecordContext(InterceptContext):
  def __init__(self, flow: MitmproxyHTTPFlow, intercept_settings: InterceptSettings):
    super().__init__(flow, intercept_settings)


from typing import TypedDict

from stoobly_agent.config.constants import alias_resolve_strategy, test_filter, test_output_level, test_strategy
from stoobly_agent.lib import logger

class TestOptions(TypedDict):
  aggregate_failures: bool
  alias_resolve_strategy: alias_resolve_strategy.AliasResolveStrategy
  assign: str
  filter: test_filter.TestFilter
  format: str
  group_by: str
  host: str
  key: str
  lifecycle_hooks_path: str
  log_level: logger.LogLevel
  output_level: test_output_level.TestOutputLevel
  public_directory_path: str
  remote_project_key: str
  report_key: str
  response_fixtures_path: str
  save: str
  scheme: str
  strategy: test_strategy.TestStrategy
  trace_id: str
  validate: str


import pdb

from typing import Tuple
from stoobly_agent.app.proxy.test.context_abc import TestContextABC as TestContext

from stoobly_agent.app.proxy.test.helpers.request_component_names_facade import RequestComponentNamesFacade
from stoobly_agent.app.proxy.test.matchers.errors import length_match_error, param_name_missing_error, type_match_error, value_match_error

from .context import MatchContext, build_match_context

def matches(test_context: TestContext, facade: RequestComponentNamesFacade, expected, actual):
    context = build_match_context(test_context, facade)

    if not context.value_type_matches(expected, actual):
        return type_match_error(context.path_key, expected, actual)

    if context.value_is_dict(actual):
        return dict_matches(context, expected, actual)
    elif context.value_is_list(actual):
        return list_matches(context, expected, actual)
    else:
        matches, log = value_matches(context, expected, actual)
        if not matches:
            return matches, 'Match error: responses do not match'
        return matches, log

def value_matches(parent_context: MatchContext, expected, actual):
    if not parent_context.value_matches(expected, actual):
        return value_match_error(parent_context.path_key, expected, actual)
    return True, ''

def dict_matches(parent_context: MatchContext, expected: dict, actual: dict) -> Tuple[bool, str]:
    if not parent_context.value_is_dict(expected) or not parent_context.value_is_dict(actual):
        return type_match_error(parent_context.path_key, dict, type(actual))

    for key, expected_value in expected.items():
        context = MatchContext(parent_context.to_dict())
        context.visit_dict(key)
        path_key = context.path_key

        # Check if key exists in actual
        if not context.param_name_exists(key, actual):
            if not context.selected() or not context.required():
                continue

            return param_name_missing_error(path_key)

        # Check if types match 
        actual_value = actual[key]
        if not context.value_fuzzy_matches(expected_value, actual_value):
            if context.ignored(expected_value, actual_value):
                continue

            return type_match_error(path_key, type(expected_value), type(actual_value))

        if context.value_is_dict(actual_value):
            matches, log = dict_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log
        elif context.value_is_list(actual_value):
            matches, log = list_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log
        else:
            # Check if value matches
            if not context.value_matches(expected_value, actual_value):
                if context.ignored(expected_value, actual_value) or not context.deterministic():
                    continue

                return value_match_error(path_key, expected_value, actual_value)

    return True, ''

def list_matches(parent_context: MatchContext, expected: list, actual: list) -> Tuple[bool, str]:
    if not parent_context.value_is_list(expected) or not parent_context.value_is_list(actual):
        return type_match_error(parent_context.path_key, list, type(actual))

    if not parent_context.length_matches(expected, actual):
        if parent_context.deterministic() and parent_context.selected():
            return length_match_error(parent_context.path_key, expected, actual)
    
    for i, expected_value in enumerate(expected):
        if i >= len(actual):
            break

        context = MatchContext(parent_context.to_dict())
        context.visit_list(i)
        path_key = context.path_key

        actual_value = actual[i]

        if not context.value_fuzzy_matches(expected_value, actual_value):
            if context.ignored(expected_value, actual_value):
                continue

            return type_match_error(path_key, type(expected_value), type(actual_value))

        if context.value_is_dict(actual_value):
            matches, log = dict_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log
        elif context.value_is_list(actual_value):
            matches, log = list_matches(context, expected_value, actual_value)
            if not matches:
                return matches, log
        else:
            if not context.value_matches(actual_value, expected_value):
                if context.ignored(expected_value, actual_value) or not context.deterministic():
                    continue

                return value_match_error(path_key, expected_value, actual_value)

    return True, ''


import pdb
import pytest

from click.testing import CliRunner

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.settings import Settings
from stoobly_agent.cli import config, record, scenario
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.config.constants import record_policy

from stoobly_agent.lib.api.keys import ProjectKey

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

class TestConfigScenarioSet():

  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestWhenRecordPolicyOverwrite():
    @pytest.fixture(scope='class')
    def created_scenario(self, runner: CliRunner):
      scenario_create_result = runner.invoke(scenario, ['create', 'test'])
      assert scenario_create_result.exit_code == 0
      _scenario = Scenario.last()
      _scenario.update(overwritable=True)
      return _scenario

    @pytest.fixture(scope='class')
    def recorded_request(self, runner: CliRunner, created_scenario: Scenario):
        record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0
        return Request.last()

    @pytest.fixture(scope='class')
    def new_scenario(self, runner: CliRunner):
      scenario_create_result = runner.invoke(scenario, ['create', 'test'])
      assert scenario_create_result.exit_code == 0

      return Scenario.last() 

    @pytest.fixture(scope='class', autouse=True)
    def change_scenario_results(self, runner: CliRunner, settings: Settings, created_scenario: Scenario, new_scenario: Scenario):
      project_key = settings.proxy.intercept.project_key
      _project_key = ProjectKey(project_key)

      data_rule = settings.proxy.data.data_rules(_project_key.id)
      data_rule.scenario_key = created_scenario.key()
      data_rule.record_policy = record_policy.OVERWRITE
      settings.commit()

      set_results = runner.invoke(config, ['scenario', 'set', new_scenario.key()])
      assert set_results.exit_code == 0

    def test_it_sets_new_scenario_overwritable(self, new_scenario: Scenario):
      _new_scenario = Scenario.find(new_scenario.id)
      assert _new_scenario.overwritable

    def test_it_sets_old_scenario_not_overwritable(self, created_scenario: Scenario):
      _created_scenario = Scenario.find(created_scenario.id)
      assert not _created_scenario.overwritable


import json
import pdb
import pytest

from click.testing import CliRunner
from typing import List
from urllib.parse import parse_qs

from stoobly_agent.test.test_helper import DETERMINISTIC_GET_REQUEST_URL, NON_DETERMINISTIC_GET_REQUEST_URL, reset

from stoobly_agent.app.models.adapters.raw_http_request_adapter import RawHttpRequestAdapter
from stoobly_agent.app.models.adapters.raw_http_response_adapter import RawHttpResponseAdapter
from stoobly_agent.app.settings.constants import firewall_action, request_component
from stoobly_agent.config.constants import custom_headers, mode, record_policy, request_origin
from stoobly_agent.cli import config, intercept, mock, record, scenario
from stoobly_agent.lib.api.keys.scenario_key import ScenarioKey
from stoobly_agent.lib.orm.request import Request
from stoobly_agent.lib.orm.response import Response
from stoobly_agent.lib.orm.scenario import Scenario
from stoobly_agent.lib.utils.decode import decode

@pytest.fixture(scope='module')
def runner():
  return CliRunner()

class TestRecording():
  @pytest.fixture(scope='class', autouse=True)
  def settings(self):
    return reset()

  class TestMocking():

    def test_it_mocks(self, runner: CliRunner):
      url = DETERMINISTIC_GET_REQUEST_URL
      record_result = runner.invoke(record, [url])
      assert record_result.exit_code == 0

      _request = Request.last()
      assert _request.url == url

      mock_result = runner.invoke(mock, [url])
      assert mock_result.exit_code == 0
      assert mock_result.stdout == record_result.stdout

  class TestRewriting():

    def test_it_checks_mode(self, runner: CliRunner):
      header_name = 'foo'
      header_value = 'bar'

      rewrite_result = runner.invoke(config, [
          'rewrite', 'set', 
          '--method', 'GET', '--mode', mode.MOCK, '--name', header_name, '--value', header_value, '--pattern', '.*?', '--type', request_component.HEADER
        ]
      )
      assert rewrite_result.exit_code == 0

      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      _request = Request.last()
      python_request = RawHttpRequestAdapter(_request.raw).to_request()

      assert python_request.headers.get(header_name.title()) == None

    class TestWhenHeaders():

      def test_it_rewrites(self, runner: CliRunner):
        header_name = 'foo'
        header_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', header_name, '--value', header_value, '--pattern', '.*?', '--type', request_component.HEADER
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _request = Request.last()
        python_request = RawHttpRequestAdapter(_request.raw).to_request()

        assert python_request.headers.get(header_name.title()) == header_value

    class TestWhenQueryParams():
    
      def test_it_rewrites(self, runner: CliRunner):
        query_param = 'foo'
        query_param_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', query_param, '--value', query_param_value, '--pattern', '.*?', '--type', request_component.QUERY_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _request = Request.last()

        query_params = parse_qs(_request.get_raw_attribute('query'))
        values = query_params.get(query_param)
        assert values != None and len(values) == 1
        assert query_param_value in values

    class TestWhenBodyParams():

      def test_it_rewrites(self, runner: CliRunner):
        body_param = 'foo'
        body_param_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', body_param, '--value', body_param_value, '--pattern', '.*?', '--type', request_component.BODY_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _request = Request.last()
        python_request = RawHttpRequestAdapter(_request.raw).to_request()

        # Since this request did not initially have a body, rewriting adds a body with header Content-Type: application/json
        assert python_request.headers.get('content-type'.title()) == 'application/json'

        body = decode(python_request.data)
        assert json.loads(body).get(body_param) == body_param_value

    class TestWhenResponseHeaders():

      def test_it_rewrites(self, runner: CliRunner):
        header_name = 'foo'
        header_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', header_name, '--value', header_value, '--pattern', '.*?', '--type', request_component.RESPONSE_HEADER
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _response = Response.last()
        python_response = RawHttpResponseAdapter(_response.raw).to_response()

        assert python_response.headers.get(header_name.title()) == header_value

    class TestWhenResponseParams():

      def test_it_rewrites(self, runner: CliRunner):
        body_param = 'foo'
        body_param_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', body_param, '--value', body_param_value, '--pattern', '.*?', '--type', request_component.RESPONSE_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _response = Response.last()
        python_response = RawHttpResponseAdapter(_response.raw).to_response()

        body = decode(python_response.content)
        assert json.loads(body).get(body_param) == body_param_value

      def test_it_does_not_rewrite(self, runner: CliRunner):
        body_param = 'foo'
        body_param_value = 'bar'

        rewrite_result = runner.invoke(config, [
            'rewrite', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--name', body_param, '--value', body_param_value, '--pattern', '.*?', '--type', request_component.RESPONSE_PARAM
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        _response = Response.last()
        python_response = RawHttpResponseAdapter(_response.raw).to_response()

        body = decode(python_response.content)
        # For non iterable resposne content, do not rewrite
        with pytest.raises(json.decoder.JSONDecodeError):
          json.loads(body)

  class TestFirewall():
    @pytest.fixture(scope='function', autouse=True)
    def settings(self):
      return reset()

    class TestWhenCli():

      def test_it_does_not_exclude(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'firewall', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--pattern', '.*?', '--action', firewall_action.EXCLUDE
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() ==  1

      def test_it_includes(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'firewall', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--pattern', DETERMINISTIC_GET_REQUEST_URL, '--action', firewall_action.INCLUDE
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 1, 'Request should be included'

        record_result = runner.invoke(record, ['--header', f"{custom_headers.REQUEST_ORIGIN}: {request_origin.WEB}", DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 2, 'Request should be included'
    
    class TestWhenNotCli():

      def test_it_excludes(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'firewall', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--pattern', '.*?', '--action', firewall_action.EXCLUDE
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, ['--header', f"{custom_headers.REQUEST_ORIGIN}: {request_origin.WEB}", DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 0

      def test_it_checks_mode(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'firewall', 'set', 
            '--method', 'GET', '--mode', mode.MOCK, '--pattern', '.*?', '--action', firewall_action.EXCLUDE
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, ['--header', f"{custom_headers.REQUEST_ORIGIN}: {request_origin.WEB}", DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 1

      def test_it_includes(self, runner: CliRunner):
        rewrite_result = runner.invoke(config, [
            'firewall', 'set', 
            '--method', 'GET', '--mode', mode.RECORD, '--pattern', DETERMINISTIC_GET_REQUEST_URL, '--action', firewall_action.INCLUDE
          ]
        )
        assert rewrite_result.exit_code == 0

        record_result = runner.invoke(record, ['--header', f"{custom_headers.REQUEST_ORIGIN}: {request_origin.WEB}", NON_DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 0, 'Request should not be included'

        record_result = runner.invoke(record, ['--header', f"{custom_headers.REQUEST_ORIGIN}: {request_origin.WEB}", DETERMINISTIC_GET_REQUEST_URL])
        assert record_result.exit_code == 0

        assert Request.count() == 1, 'Request should be included'

  class TestScenario():

    @pytest.fixture()
    def scenario_key(self, runner: CliRunner):
      res = runner.invoke(scenario, ['create', '--select', 'key', '--without-headers', 'test-scenario'])
      assert res.exit_code == 0
      return ScenarioKey(res.stdout.strip())

    def test_it_sets_scenario_id(self, runner: CliRunner, scenario_key: ScenarioKey):
      record_result = runner.invoke(record, ['--scenario-key', scenario_key.raw, DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      _request = Request.last()
      assert _request.scenario.uuid == scenario_key.id

  class TestNotFoundPolicy():

    def test_it_records_only_not_found_requests(self, runner: CliRunner):
      intercept_result = runner.invoke(intercept, ['configure', '--policy', record_policy.NOT_FOUND])
      assert intercept_result.exit_code == 0

      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 1

      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 1

      record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 2

  class TestFoundPolicy():
    @pytest.fixture(scope='function', autouse=True)
    def settings(self):
      return reset()

    def test_it_does_not_record_any_requests(self, runner: CliRunner):
      intercept_result = runner.invoke(intercept, ['configure', '--policy', record_policy.FOUND])
      assert intercept_result.exit_code == 0

      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 0

      record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 0

    def test_it_records_only_found_requests(self, runner: CliRunner):
      record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 1

      intercept_result = runner.invoke(intercept, ['configure', '--policy', record_policy.FOUND])
      assert intercept_result.exit_code == 0

      record_result = runner.invoke(record, [DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 1

      record_result = runner.invoke(record, [NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      assert Request.count() == 2

  class TestOverwritePolicy():

    @pytest.fixture(scope='class', autouse=True)
    def created_scenario(self, runner: CliRunner):
      scenario_create_result = runner.invoke(scenario, ['create', 'test'])
      assert scenario_create_result.exit_code == 0

      return Scenario.last()

    @pytest.fixture(scope='class', autouse=True)
    def recorded_requests(self, runner: CliRunner, created_scenario: Scenario):
      record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0
      return Request.all()

    @pytest.fixture(scope='class', autouse=True)
    def enable_overwrite_results(self, runner: CliRunner, created_scenario: Scenario, recorded_requests: List[Request]):
      _created_scenario = Scenario.find(created_scenario.id)
      assert _created_scenario.requests_count == len(recorded_requests)

      config_results = runner.invoke(intercept, ['configure', '--mode', mode.RECORD, '--policy', record_policy.OVERWRITE])
      assert config_results.exit_code == 0

      set_results = runner.invoke(config, ['scenario', 'set', created_scenario.key()])
      assert set_results.exit_code == 0
      return config_results

    @pytest.fixture(scope='class')
    def recorded_new_requests(self, runner: CliRunner, created_scenario: Scenario, recorded_requests: List[Request]):
      _created_scenario = Scenario.find(created_scenario.id)
      assert _created_scenario.requests_count == len(recorded_requests)
      assert _created_scenario.overwritable

      # Only on first request record do overwrite the scenario
      record_result = runner.invoke(record, ['--scenario-key', created_scenario.key(), NON_DETERMINISTIC_GET_REQUEST_URL])
      assert record_result.exit_code == 0

      _created_scenario = Scenario.find(created_scenario.id)
      return _created_scenario.requests

    def test_it_sets_scenario_not_overwritable(self, created_scenario: Scenario, recorded_new_requests: List[Request]):
      _created_scenario = Scenario.find(created_scenario.id)
      assert _created_scenario.requests_count == len(recorded_new_requests)
      assert _created_scenario.overwritable == False


import pdb
import pytest

from stoobly_agent.app.cli.scaffold.hosts_file_manager import HostsFileManager


class TestHostsFileManager():

  @pytest.fixture
  def hosts_file_manager(self):
    yield HostsFileManager()

  def test_get_hosts_file_path(self, hosts_file_manager):
    hosts_file_path = hosts_file_manager.get_hosts_file_path()

    # Test runners are all Linux distros for now
    assert hosts_file_path == '/etc/hosts'

  def test_get_hosts(self, hosts_file_manager):
    hosts = hosts_file_manager.get_hosts()

    assert hosts
    assert len(hosts) > 1

    localhost_found = False
    for host in hosts:
      if host.ip_address == '127.0.0.1' and 'localhost' in host.hostnames:
        localhost_found = True
        break
    assert localhost_found

  def test_find_host(self, hosts_file_manager):
    url = 'localhost'

    host = hosts_file_manager.find_host(url)

    assert host
    assert host.ip_address == '127.0.0.1'
    assert 'localhost' in host.hostnames


  class TestsSplitHostsLine():
    def test_basic(self, hosts_file_manager):
      line = '0.0.0.0 example.com'
      split = hosts_file_manager._HostsFileManager__split_hosts_line(line)

      assert split[0] == '0.0.0.0'
      assert split[1] == 'example.com'

    def test_tabs(self, hosts_file_manager):
      line = '0.0.0.0\texample.com'
      split = hosts_file_manager._HostsFileManager__split_hosts_line(line)

      assert split[0] == '0.0.0.0'
      assert split[1] == 'example.com'

    def test_comment(self, hosts_file_manager):
      line = '# 0.0.0.0 example.com'
      split = hosts_file_manager._HostsFileManager__split_hosts_line(line)

      assert len(split) == 0

    def test_inline_comment(self, hosts_file_manager):
      line = '0.0.0.0\texample.com # Comment'
      split = hosts_file_manager._HostsFileManager__split_hosts_line(line)

      assert split[0] == '0.0.0.0'
      assert split[1] == 'example.com'

    def test_multiple_hostnames(self, hosts_file_manager):
      line = '0.0.0.0\texample.com example2.com example3.com # Comment'
      split = hosts_file_manager._HostsFileManager__split_hosts_line(line)

      assert split[0] == '0.0.0.0'
      assert split[1] == 'example.com'
      assert split[2] == 'example2.com'
      assert split[3] == 'example3.com'





import pdb

from requests.structures import CaseInsensitiveDict

from ..replay.body_parser_service import decode_response, encode_response

class TestContextResponse():

  def __init__(self):
    self._content = ''
    self._headers = CaseInsensitiveDict()
    self._status_code = None

  @property
  def content(self):
    return self._content  

  @property
  def content_type(self):
    return self.headers.get('content-type')

  @property
  def headers(self):
    return self._headers

  @property
  def status_code(self):
    return self._status_code

  def decode_content(self):
    content_type = self.content_type
    return decode_response(self.content, content_type)

  def encode_content(self):
    content_type = self.content_type
    return encode_response(self.content, content_type)

  def with_content(self, content):
    self._content = content
    return self

  def with_headers(self, headers):
    _headers = CaseInsensitiveDict()

    for name in headers:
      _headers[name] = headers[name]

    self._headers = _headers
    return self

  def with_status_code(self, status_code):
    self._status_code = status_code
    return self



import os
import shutil

from stoobly_agent.config.data_dir import DataDir, DATA_DIR_NAME

class App():

  def __init__(self, path: str, scaffold_namespace: str, **kwargs):
    path = os.path.abspath(path) or os.getcwd()
    data_dir: DataDir = DataDir.instance(path) 

    self.__data_dir_path = data_dir.path
    self.__ca_certs_dir_path = kwargs.get('ca_certs_dir_path') or data_dir.ca_certs_dir_path
    self.__certs_dir_path = kwargs.get('certs_dir_path') or data_dir.certs_dir_path
    self.__context_dir_path = kwargs.get('context_dir_path') or data_dir.context_dir_path
    self.__data_dir = data_dir
    self.__dir_path = path
    self.__scaffold_namespace = scaffold_namespace
    self.__skip_validate_path = not not kwargs.get('dry_run')

  @property
  def ca_certs_dir_path(self):
    return self.__ca_certs_dir_path

  @ca_certs_dir_path.setter
  def ca_certs_dir_path(self, v: str):
    self.__validate_path(v)
    self.__certs_dir_path = v

  @property
  def certs_dir_path(self):
    return self.__certs_dir_path

  @certs_dir_path.setter
  def certs_dir_path(self, v: str):
    self.__validate_path(v)
    self.__certs_dir_path = v

  @property
  def context_dir_path(self):
    return self.__context_dir_path

  @context_dir_path.setter
  def context_dir_path(self, v: str):
    self.__validate_path(v)
    self.__context_dir_path = v 

  @property
  def data_dir(self):
    return self.__data_dir

  @property
  def data_dir_path(self):
    return os.path.join(self.context_dir_path, DATA_DIR_NAME)

  @property
  def valid(self):
    return os.path.exists(self.scaffold_namespace_path)

  @property
  def scaffold_namespace(self):
    return self.__scaffold_namespace

  @property
  def scaffold_namespace_path(self):
    return os.path.join(self.data_dir_path, self.scaffold_namespace)

  @property
  def dir_path(self):
    return self.__dir_path

  @property
  def data_dir_path(self):
    return self.__data_dir_path

  @property
  def scaffold_namespace_path(self):
    return os.path.join(self.data_dir_path, self.scaffold_namespace)

  @property
  def services(self):
    return list(map(lambda path: os.path.basename(path), self.service_paths))

  @property
  def service_paths(self):
    services_dir = os.path.join(self.data_dir_path, self.scaffold_namespace)

    services = []
    for filename in os.listdir(services_dir):
      path = os.path.join(services_dir, filename)
      if not os.path.isdir(path):
        continue
      
      services.append(path)

    return services

  def copy_folders_and_hidden_files(self, src, dst):
      os.makedirs(dst, exist_ok=True)

      # Walk through the source directory
      for root, dirs, files in os.walk(src):
          # Copy hidden files only
          for file_name in files:
              src_file_path = os.path.join(root, file_name)
              dst_file_path = os.path.join(dst, os.path.relpath(root, src), file_name)

              if not file_name.startswith('.'):
                  if os.path.exists(dst_file_path):
                      continue

              os.makedirs(os.path.dirname(dst_file_path), exist_ok=True)  # Create directories in destination
              shutil.copy2(src_file_path, dst_file_path)

  def __validate_path(self, v: str):
    if not isinstance(v, str):
      raise TypeError('Expected a str')

    if not self.__skip_validate_path:
      if not os.path.exists(v):
        raise ValueError(f"{v} does not exist")


import os
import pdb
import socket
import time

from collections import Counter
from docker import errors as docker_errors
from pathlib import Path

import yaml
from docker.models.containers import Container

from stoobly_agent.app.cli.scaffold.constants import (
  PUBLIC_FOLDER_NAME,
  STOOBLY_DATA_DIR,
  WORKFLOW_RECORD_TYPE,
  WORKFLOW_TEST_TYPE,
)
from stoobly_agent.app.cli.scaffold.docker.constants import APP_EGRESS_NETWORK_TEMPLATE
from stoobly_agent.app.cli.scaffold.hosts_file_manager import HostsFileManager
from stoobly_agent.app.cli.scaffold.service_command import ServiceCommand
from stoobly_agent.app.cli.scaffold.service_docker_compose import ServiceDockerCompose
from stoobly_agent.app.cli.scaffold.validate_command import ValidateCommand
from stoobly_agent.app.cli.scaffold.validate_exceptions import ScaffoldValidateException
from stoobly_agent.config.data_dir import DATA_DIR_NAME
from stoobly_agent.lib.logger import bcolors

from .app import App


class ServiceWorkflowValidateCommand(ServiceCommand, ValidateCommand):
  def __init__(self, app: App, **kwargs):
    ServiceCommand.__init__(self, app, **kwargs)
    ValidateCommand.__init__(self)

    self.workflow_name = kwargs['workflow_name']
    self.hostname = self.service_config.hostname
    self.service_docker_compose = ServiceDockerCompose(
      app_dir_path=app.dir_path, target_workflow_name=self.workflow_name, service_name=self.service_name, hostname=self.hostname
    )

  @property
  def public_dir_path(self):
    return os.path.join(self.workflow_path, PUBLIC_FOLDER_NAME)
  
  @property
  def workflow_path(self):
    return os.path.join(
      self.data_dir_path,
      self.workflow_relative_path
    )
  @property
  def workflow_relative_path(self):
    return os.path.join(
      self.service_relative_path,
      self.workflow_name
    )

  def is_local(self):
    with open (self.service_docker_compose.docker_compose_path,'rb') as f:
      docker_compose_file_content = yaml.safe_load(f)
      if docker_compose_file_content and docker_compose_file_content.get('services'):
        return True

      # We can potentially check the port too someday

    return False

  def is_external(self):
    return not self.is_local()

  def hostname_reachable(self, hostname: str, port: int) -> bool:
    print(f"Validating connection to hostname: {hostname}, port: {port}")
    timeout = 1
    retries = 15
    delay = 0.100

    for attempt in range(retries):
      try:
        with socket.create_connection((hostname, port), timeout=timeout):
          return True
      except (socket.timeout, socket.error):
        if attempt < retries - 1:
          time.sleep(delay)

    hostname_not_reachable_message = f"{bcolors.FAIL}Connection failed to hostname: {hostname}, port: {port}, num_retries: {retries}{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Try confirming that {hostname} is reachable from your machine or environment.{bcolors.ENDC}"
    error_message = f"{hostname_not_reachable_message}\n\n{suggestion_message}"
    raise ScaffoldValidateException(error_message)

  # Check if hostname is defined in hosts file
  def hostname_exists(self, hostname: str) -> bool:
    print(f"Validating hostname exists in hosts file for hostname: {hostname}")

    hosts_file_manager = HostsFileManager()
    host_mapping = hosts_file_manager.find_host(hostname)
    if host_mapping:
      print(f"Correct hosts mapping found for {hostname}")
      return True

    hosts_file_path = hosts_file_manager.get_hosts_file_path()

    missing_host_message = f"{bcolors.FAIL}Missing hosts mapping for {hostname}{bcolors.ENDC}"
    suggestion_message = f"{bcolors.BOLD}Confirm hostname '{hostname}' is in your hosts file at '{hosts_file_path}'. If not there, please add it{bcolors.ENDC}"
    error_message = f"{missing_host_message}\n\n{suggestion_message}"
    raise ScaffoldValidateException(error_message)
 
  def validate_hostname(self, hostname: str, port: int) -> None:
    print(f"Validating hostname: {hostname}")

    self.hostname_exists(hostname)

    self.hostname_reachable(hostname, port)

    # TODO: check logs of proxy. lifecycle hook for custom logging? Does mitmproxy support json logging?

  def validate_internal_hostname(self, url: str) -> None:
    print(f"Validating hostname inside Docker network, url: {url}")
    
    timeout_seconds = 1
    output = self.docker_client.containers.run(
      image='curlimages/curl:8.11.0',
      command=f"curl --max-time {timeout_seconds} {url} --verbose",
      network=APP_EGRESS_NETWORK_TEMPLATE.format(network=self.app_config.network),
      stderr=True,
      remove=True,
    )

    # Note: 499 error could also mean success because it shows the proxy
    # connection is working, but we haven't recorded anything yet
    logs = output.decode('ascii')
    if ('200 OK' not in logs) and ('499' not in logs):
      raise ScaffoldValidateException(f"Error reaching {url} from inside Docker network")

  # Check public folder exists in container
  def validate_public_folder(self, container: Container):
    
    if self.workflow_name == WORKFLOW_RECORD_TYPE:
      print(f"Skipping validating public folder in workflow: {self.workflow_name}, container: {container.name}")
      return

    print(f"Validating public folder in container: {container.name}")

    data_dir_mounted = False
    volume_mounts = container.attrs['Mounts']

    for volume_mount in volume_mounts:
      if volume_mount['Destination'] == STOOBLY_DATA_DIR:
        data_dir_mounted = True
        break
    if not data_dir_mounted:
      raise ScaffoldValidateException(f"Data directory is not mounted for: {container.name}")

    # Only the running proxy containers will be checkable
    if container.status == 'exited':
      print(f"Skipping validating public folder contents because container is exited: {container.name}")
      return

    # Check contents of public folder to confirm it's shared
    public_folder_path = f"{PUBLIC_FOLDER_NAME}"
    exec_result = container.exec_run(f"ls -A {public_folder_path}")
    output = exec_result.output

    public_folder_contents_container = output.decode('ascii').split('\n')
    if public_folder_contents_container[-1] == '':
      public_folder_contents_container.pop()
    public_folder_contents_scaffold = os.listdir(self.public_dir_path)

    if Counter(public_folder_contents_container) != Counter(public_folder_contents_scaffold):
      raise ScaffoldValidateException(f"public folder was not mounted properly, expected {self.public_dir_path} to exist in container path {public_folder_path}")

  def validate_proxy_container(self, service_proxy_container: Container):
    print(f"Validating proxy container: {service_proxy_container.name}")

    if service_proxy_container.status == 'exited' or service_proxy_container.attrs['State']['ExitCode'] != 0:
      raise ScaffoldValidateException(f"Proxy container is exited: {service_proxy_container.name}")
    if service_proxy_container.status != 'running':
      raise ScaffoldValidateException(f"Proxy container is not running: {service_proxy_container.name}")
    if not service_proxy_container.attrs:
      raise ScaffoldValidateException(f"Container attributes are missing for: {service_proxy_container.name}")

    if not self.service_config.detached:
      self.validate_public_folder(service_proxy_container)

  def validate_service_container(self):
    pass
  
  def validate(self) -> bool:
    print(f"Validating service: {self.service_name}")

    url = f"{self.service_config.scheme}://{self.hostname}"

    if self.service_config.hostname and self.workflow_name not in [WORKFLOW_TEST_TYPE]:
      self.validate_hostname(self.hostname, self.service_config.port)
    
    # Test workflow won't expose services that are detached and have a hostname to the host such as assets.
    # Need to test connection from inside the Docker network
    if self.service_config.hostname and self.workflow_name == WORKFLOW_TEST_TYPE and self.service_config.detached:
      self.validate_internal_hostname(url)

    self.validate_init_containers(self.service_docker_compose.init_container_name, self.service_docker_compose.configure_container_name)

    # Service init containers have a mounted dist folder unlike the core init container
    init_container = self.docker_client.containers.get(self.service_docker_compose.init_container_name)
    self.validate_public_folder(init_container)

    if self.service_config.hostname:
      try:
        service_proxy_container = self.docker_client.containers.get(self.service_docker_compose.proxy_container_name)
      except docker_errors.NotFound:
        error_message = self._ValidateCommand__generate_container_not_found_error(self.service_docker_compose.proxy_container_name)
        raise ScaffoldValidateException(error_message)

      self.validate_proxy_container(service_proxy_container)

    # External services won't have a container to check
    if self.is_local() or self.service_config.detached:
      container_name = self.service_docker_compose.container_name
      try:
        service_container = self.docker_client.containers.get(container_name)
      except docker_errors.NotFound:
        error_message = self._ValidateCommand__generate_container_not_found_error(container_name)
        raise ScaffoldValidateException(error_message)
      if service_container.status == 'exited' or service_container.attrs['State']['ExitCode'] != 0:
        raise ScaffoldValidateException(f"Custom container is exited: {service_container.name}")
      if service_container.status != 'running':
        raise ScaffoldValidateException(f"Custom container is not running: {service_container.name}")
    
    if self.is_local():
      print(f"Validating local user defined service: {self.service_name}")
      # Validate docker-compose path exists
      docker_compose_path = f"{self.app_dir_path}/{DATA_DIR_NAME}/docker/{self.service_docker_compose.service_name}/{self.workflow_name}/docker-compose.yml"
      destination_path = Path(docker_compose_path)

      if not destination_path.exists():
        message = f"{bcolors.FAIL}Docker Compose file does not exist: {destination_path}{bcolors.ENDC}"
        suggestion_message = f"{bcolors.BOLD}A missing Docker Compose file often means it got deleted by accident. Please restore from backup or rerun the scaffold service create command.{bcolors.ENDC}"
        error_message = f"{message}\n\n{suggestion_message}"
        raise ScaffoldValidateException(error_message)
      if not destination_path.is_file():
        raise ScaffoldValidateException(f"Docker compose path is not a file: {destination_path}")

      # Validate docker-compose.yml file has the service defined
      with open(destination_path) as f:
        if self.service_name not in f.read():
          message = f"{bcolors.FAIL}Custom container service is not defined in Docker Compose file: {destination_path}{bcolors.ENDC}"
          suggestion_message = f"{bcolors.BOLD}Please add your service definition. See https://docs.docker.com/reference/compose-file/services {bcolors.ENDC}"
          error_message = f"{message}\n\n{suggestion_message}"
          raise ScaffoldValidateException(error_message)

    if self.service_config.detached:
      self.validate_detached(service_container)

    print(f"{bcolors.OKGREEN}✔ Done validating service: {self.service_name}, success!{bcolors.ENDC}\n")

    return True




import distro
import os
import pdb
import socket
import ssl
import subprocess

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.x509.oid import NameOID
from mitmproxy.certs import CertStore
from pathlib import Path

from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.logger import Logger

LOG_ID = 'CertificateAuthority'

MITMPROXY_CN = 'mitmproxy'

class CertificateAuthority():

    def __init__(self, certs_dir = DataDir.instance().ca_certs_dir_path, cn = MITMPROXY_CN):
        self.certs_dir = certs_dir
        self.cn = cn
        self.key_size = 2048

    @property
    def certs_generated(self):
        if not os.path.exists(self.ca_cert_path('.pem')):
            return False

        if not os.path.exists(self.ca_cert_path('.cer')):
            return False

        if not os.path.exists(self.ca_path('.pem')):
            return False

        return True

    def ca_file_name(self, extension: str):
        return '{cn}-ca{extension}'.format(cn=self.cn, extension=extension)

    def ca_path(self, extension: str):
        file_name = self.ca_file_name(extension)
        return os.path.join(self.certs_dir, file_name)

    def ca_cert_file_name(self, extension: str):
        return '{cn}-ca-cert{extension}'.format(cn=self.cn, extension=extension)

    def ca_cert_path(self, extension: str):
        file_name = self.ca_cert_file_name(extension)
        return os.path.join(self.certs_dir, file_name)

    def generate_certs(self):
        if not os.path.exists(self.certs_dir):
            os.makedirs(self.certs_dir, exist_ok=True)

        path = Path(self.certs_dir)
        CertStore.create_store(path, self.cn, self.key_size)

    # https://askubuntu.com/a/94861
    def handle_debian(self):
        extra_ca_certs_dir = '/usr/local/share/ca-certificates/extra'

        ca_cert_path = self.ca_cert_path('.cer')
        self.__ensure_exists(ca_cert_path)

        extra_crt_absolute_path = os.path.join(extra_ca_certs_dir, self.ca_cert_file_name('.crt'))

        subprocess.run(f"sudo mkdir -p {extra_ca_certs_dir}".split(), check=True)
        subprocess.run(f"sudo cp {ca_cert_path} {extra_crt_absolute_path}".split(), check=True)
        subprocess.run("sudo update-ca-certificates".split(), check=True)

    # https://www.dssw.co.uk/reference/security.html
    def handle_darwin(self):
        system_keychain_path = '/Library/Keychains/System.keychain'

        ca_cert_path = self.ca_cert_path('.pem')
        self.__ensure_exists(ca_cert_path)

        subprocess.run(f"sudo security add-trusted-cert -d -p ssl -p basic -k {system_keychain_path} {ca_cert_path}".split(), check=True)

    def handle_rhel(self):
        ca_trust_dir = '/etc/pki/ca-trust/source/anchors'

        ca_cert_path = self.ca_cert_path('.cer')
        self.__ensure_exists(ca_cert_path)

        ca_trust_crt_absolute_path = os.path.join(ca_trust_dir, self.crt_file_name)

        subprocess.run(f"sudo cp {ca_cert_path} {ca_trust_crt_absolute_path}".split(), check=True)
        subprocess.run("sudo update-ca-trust extract".split(), check=True)

    def install(self):
        if not self.certs_generated:
            self.generate_certs()

        distro_name = distro.name(pretty=True)

        # Ubuntu or other Debian based
        if distro.like() == 'debian':
            self.handle_debian()
        # MacOS
        elif distro.id() == 'darwin':
            self.handle_darwin()
        # elif distro.id() == 'rhel':
        #     return
        else:
            raise TypeError(
                f"{distro_name} is not supported yet for automatic installation. For manual installation, see https://docs.mitmproxy.org/stable/concepts-certificates/"
            )

    def sign(self, hostname: str, dest = None, port = 443):
        dest = dest or self.certs_dir

        if self.signed(hostname, dest):
            return False

        store = CertStore.from_store(self.certs_dir, self.cn, self.key_size)
        organization, alt_names = self.__cert_details(hostname, port)
        cert_entry = store.get_cert(alt_names[0], alt_names, organization)

        crt_path = os.path.join(dest, f"{hostname}.crt")
        with open(crt_path, 'wb') as fp:
            cert = cert_entry.cert
            crt = cert.to_pem()
            fp.write(crt)

        key_path = os.path.join(dest, f"{hostname}.key")
        with open(key_path, 'wb') as fp:
            private_key = cert_entry.privatekey
            key = private_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
            fp.write(key)

        joined_path = os.path.join(dest, f"{hostname}-joined.pem")
        with open(joined_path, 'wb') as fp:
            fp.write(b"\n".join([crt, key]))

        return True

    def signed(self, hostname: str, dest = None):
        dest = dest or self.certs_dir

        crt_path = os.path.join(dest, f"{hostname}.crt")
        if not os.path.exists(crt_path):
            return False

        key_path = os.path.join(dest, f"{hostname}.key")
        if not os.path.exists(key_path):
            return False

        joined_path = os.path.join(dest, f"{hostname}-joined.pem")
        if not os.path.exists(joined_path):
            return False

        return True

    def __ensure_exists(self, file_path):
        if not os.path.exists(file_path):
            self.generate_certs()

    def __cert_details(self, domain: str, port: int = 443):
        """
        Retrieves the organization name and subject alternative names from an HTTPS domain's certificate.

        :param domain: The domain name (e.g., 'example.com').
        :param port: The port number (default: 443 for HTTPS).
        :return: A tuple containing the organization name and a list of alternative names (or None if not found).
        """
        alt_names = [domain]
        try:
            # Connect to the server and get the certificate
            context = ssl.create_default_context()
            context.minimum_version = ssl.TLSVersion.TLSv1_2
            timeout = 2.5
            with socket.create_connection((domain, port), timeout) as sock:
                with context.wrap_socket(sock, server_hostname=domain) as ssock:
                    cert_der = ssock.getpeercert(binary_form=True)

            # Parse the certificate
            cert = x509.load_der_x509_certificate(cert_der, default_backend())

            # Extract the organization name
            try:
                org_name = cert.subject.get_attributes_for_oid(NameOID.ORGANIZATION_NAME)[0].value
            except IndexError:
                org_name = None

            # Extract the Subject Alternative Names (SANs)
            try:
                ext = cert.extensions.get_extension_for_class(x509.SubjectAlternativeName)
                alt_names.extend(ext.value.get_values_for_type(x509.DNSName))
            except x509.ExtensionNotFound:
                alt_names = []

            return org_name, alt_names

        except Exception as e:
            Logger.instance(LOG_ID).debug(f"Could not retrieve certificate for {domain}: {e}")
            return None, alt_names


from stoobly_agent.app.proxy.test.context_abc import TestContextABC as TestContext

def matches(context: TestContext):
    lifecycle_hooks = context.lifecycle_hooks

    if not context.lifecycle_hooks_path:
        return False, 'Missing lifecycle hooks path' 

    if not 'handle_test' in lifecycle_hooks:
        return False, f"Expected function 'handle_test' to be defined in {context.lifecycle_hooks_path}"

    try:
        status, log = lifecycle_hooks['handle_test'](context)
    except Exception as e:
        return False, f"Exception: {e}"

    if not type(status) is bool or not type(log) is str:
        return False, f"Expected function 'test' to return [bool, str], got [{type(status)}, {type(log)}]"

    return status, log


import os
import pdb
import pytest
import requests

from click.testing import CliRunner
from mitmproxy.http import Request as MitmproxyRequest

from stoobly_agent.test.test_helper import reset

from stoobly_agent.app.models.adapters.orm.request.mitmproxy_adapter import MitmproxyRequestAdapter
from stoobly_agent.app.models.factories.resource.local_db.helpers.request_builder import RequestBuilder
from stoobly_agent.app.proxy.mock.eval_fixtures_service import eval_fixtures
from stoobly_agent.app.proxy.mock.types import Fixtures
from stoobly_agent.app.settings import Settings
from stoobly_agent.config.data_dir import DataDir
from stoobly_agent.lib.orm.request import Request

@pytest.fixture(scope='module')
def runner():
    return CliRunner()

@pytest.fixture(scope='module')
def settings():
  return reset()

class TestEvalFixturesService():
  class TestResponseFixtures():
    @pytest.fixture(scope='class')
    def request_method(self):
      return 'POST'

    @pytest.fixture(scope='class')
    def request_url(self):
      return 'https://petstore.swagger.io/404.html'

    @pytest.fixture(scope='class')
    def created_request(
      self, settings: Settings, request_method: str, request_url: str
    ):
      status = RequestBuilder(
        method=request_method,
        response_body='',
        status_code=200,
        url=request_url,
      ).with_settings(settings).build()[1]
      assert status == 200

      return Request.last()

    @pytest.fixture(scope='class')
    def not_found_file_contents(self):
      return b'Not Found'

    @pytest.fixture(scope='class')
    def public_directory(self):
      tmp_dir_path = DataDir.instance().tmp_dir_path
      public_dir_path = os.path.join(tmp_dir_path, 'public') 
      if not os.path.exists(public_dir_path):
        os.mkdir(public_dir_path)
      return public_dir_path

    @pytest.fixture(autouse=True, scope='class')
    def not_found_file_path(self, public_directory: str, not_found_file_contents: str):
      path = os.path.join(public_directory, '404.html')
      with open(path, 'wb') as fp:
        fp.write(not_found_file_contents)
      return path

    @pytest.fixture(scope='class')
    def response_fixtures(self, request_method: str, not_found_file_path: str) -> Fixtures:
      fixtures = {}
      fixtures[request_method] = {
        '/404.html': {
          'headers': {
            'test': '1',
          },
          'path': not_found_file_path,
          'status_code': 404,
        }
      }
      return fixtures

    @pytest.fixture()
    def mitmproxy_request(self, created_request: Request) -> MitmproxyRequest:
      return MitmproxyRequestAdapter(created_request).adapt()

    @pytest.fixture()
    def fixtures_response(self, mitmproxy_request: MitmproxyRequest, response_fixtures: Fixtures):
      res: requests.Response = eval_fixtures(mitmproxy_request, response_fixtures=response_fixtures)
      assert res != None
      return res

    def test_it_sets_response(
      self, fixtures_response: requests.Response, not_found_file_contents: str
    ):
      assert fixtures_response.raw.read() == not_found_file_contents

    def test_it_sets_headers(self, fixtures_response: requests.Response):
      assert fixtures_response.headers['test'] == '1'
      assert fixtures_response.headers['Content-Type'] == 'text/html'

    def test_it_sets_status_code(self,  fixtures_response: requests.Response):
      assert fixtures_response.status_code == 404

  class TestPublicDirectory():
    @pytest.fixture(scope='class')
    def request_method(self):
      return 'GET'

    @pytest.fixture(scope='class')
    def request_url(self):
      return 'https://petstore.swagger.io'

    @pytest.fixture(scope='class')
    def created_request(
      self, settings: Settings, request_method: str, request_url: str
    ):
      status = RequestBuilder(
        method=request_method,
        request_headers={'accept': 'text/html;q=0.1,application/json;q=0.9'},
        response_body='',
        status_code=200,
        url=request_url,
      ).with_settings(settings).build()[1]
      assert status == 200

      return Request.last()

    @pytest.fixture()
    def mitmproxy_request(self, created_request: Request) -> MitmproxyRequest:
      return MitmproxyRequestAdapter(created_request).adapt()

    @pytest.fixture(scope='class')
    def public_directory(self):
      tmp_dir_path = DataDir.instance().tmp_dir_path
      public_dir_path = os.path.join(tmp_dir_path, 'public') 
      if not os.path.exists(public_dir_path):
        os.mkdir(public_dir_path)
      return public_dir_path

    @pytest.fixture(scope='class')
    def index_file_contents(self):
      return b'Hello World!'

    @pytest.fixture(autouse=True, scope='class')
    def index_file_path(self, public_directory: str, index_file_contents: str):
      path = os.path.join(public_directory, 'index.html')
      with open(path, 'wb') as fp:
        fp.write(index_file_contents)
      return path

    @pytest.fixture()
    def public_directory_response(self, mitmproxy_request: MitmproxyRequest, public_directory: str):
      res: requests.Response = eval_fixtures(mitmproxy_request, public_directory_path=public_directory)
      assert res != None
      return res

    def test_it_sets_contents(
      self, public_directory_response: requests.Response, index_file_contents: str
    ):
      assert public_directory_response.raw.read() == index_file_contents

    def test_it_headers(self, public_directory_response: requests.Response):
      assert public_directory_response.headers['Content-Type'] == 'text/html'
    
    def test_it_sets_status_code(self,  public_directory_response: requests.Response):
      assert public_directory_response.status_code == 200